"""
Mapping of the R library "grDevices" for graphical devices
"""

import io
import os
import tempfile
from contextlib import contextmanager
from rpy2.robjects.packages import importr, WeakPackage

grdevices = importr('grDevices')

grdevices = WeakPackage(grdevices._env,
                        grdevices.__rname__,
                        translation=grdevices._translation,
                        exported_names=grdevices._exported_names,
                        on_conflict="warn",
                        version=grdevices.__version__,
                        symbol_r2python=grdevices._symbol_r2python,
                        symbol_resolve=grdevices._symbol_resolve)

# non-interactive file devices
png = grdevices.png
jpeg = grdevices.jpeg
bmp = grdevices.bmp
tiff = grdevices.tiff

svg = grdevices.svg

postscript = grdevices.postscript
pdf = grdevices.pdf
cairo_pdf = grdevices.cairo_pdf

#
X11 = grdevices.X11

# OS X
quartz = grdevices.quartz

# devices
dev_list = grdevices.dev_list
dev_cur = grdevices.dev_cur
dev_next = grdevices.dev_next
dev_prev = grdevices.dev_prev
dev_set = grdevices.dev_set
dev_off = grdevices.dev_off


@contextmanager
def render_to_file(device, *device_args, **device_kwargs):
    """
    Context manager that returns a R figures in a file object.

    :param device: an R "device" function. This function is expected
                   to take a filename as its first argument.

    """
    # TODO: better function signature to check that a file name is passed.
    current = dev_cur()[0]
    try:
        device(*device_args, **device_kwargs)
        yield None
    finally:
        if current != dev_cur()[0]:
            dev_off()


@contextmanager
def render_to_bytesio(device, *device_args, **device_kwargs):
    """
    Context manager that returns a R figures in a :class:`io.BytesIO`
    object.

    :param device: an R "device" function. This function is expected
                   to take a filename as its first argument.

    """
    fn = tempfile.mktemp()
    b = io.BytesIO()
    current = dev_cur()[0]
    try:
        device(fn, *device_args, **device_kwargs)
        yield b
    finally:
        if current != dev_cur()[0]:
            dev_off()
        if os.path.exists(fn):
            with open(fn, 'rb') as fh:
                b.write(fh.read())
            os.unlink(fn)



"""
Wrapper for the popular R library ggplot2.

With rpy2, the most convenient general way to
import packages is to use `importr()`, for example
with ggplot2::

    from rpy2.robjects.packages import importr
    ggplot2 = importr('ggplot2')

This module is an supplementary layer in which an attempt
at modelling the package as it was really developed
as Python package is made. Behind the scene, `importr()`
is used and can be accessed with:

    from robjects.robjects.lib import ggplot2
    ggplot2.ggplot2

GGplot2 is designed using a prototype-based approach to
Object-Oriented Programming, and this module is trying
to define class-hierachies so the nature of a given
instance can be identified more easily.

The main families of classes are:

- GGplot
- Aes and AesString
- Layer
- Stat

A downside of the approach is that the code in the
module is 'hand-made'. In hindsight, this can be tedious
to maintain and document but this is a good showcase of
"manual" mapping of R code into Python classes.

The codebase in R for ggplot2 has evolved since this
was initially written, and many functions have
signature-defined parameters (used to be ellipsis
about everywhere). Metaprogramming will hopefully
be added to shorten the Python code in the module,
and provide a more dynamic mapping.

"""

import rpy2.robjects as robjects
import rpy2.robjects.constants
import rpy2.robjects.conversion as conversion
from rpy2.robjects.packages import importr, WeakPackage
from rpy2.robjects import rl
import warnings

NULL = robjects.NULL

rlang = importr('rlang', on_conflict='warn',
                robject_translations={'.env': '__env'})
lazyeval = importr('lazyeval', on_conflict='warn')
base = importr('base', on_conflict='warn')
ggplot2 = importr('ggplot2', on_conflict='warn')
ggplot2 = WeakPackage(ggplot2._env,
                      ggplot2.__rname__,
                      translation=ggplot2._translation,
                      exported_names=ggplot2._exported_names,
                      on_conflict="warn",
                      version=ggplot2.__version__,
                      symbol_r2python=ggplot2._symbol_r2python,
                      symbol_resolve=ggplot2._symbol_resolve)

TARGET_VERSION = '3.3.'
if not ggplot2.__version__.startswith(TARGET_VERSION):
    warnings.warn(
        'This was designed againt ggplot2 versions starting with %s but you '
        'have %s' % (TARGET_VERSION, ggplot2.__version__))
ggplot2_env = robjects.baseenv['as.environment']('package:ggplot2')

StrVector = robjects.StrVector


def as_symbol(x):
    return rlang.sym(x)


_AES_RLANG = rl('ggplot2::aes()')


class GGPlot(robjects.vectors.ListVector):
    """ A Grammar of Graphics Plot.

    GGPlot instances can be added to one an other in order to construct
    the final plot (the method `__add__()` is implemented).
    """

    _constructor = ggplot2._env['ggplot']
    _rprint = ggplot2._env['print.ggplot']
    _add = ggplot2._env['%+%']

    @classmethod
    def new(cls, data, mapping=_AES_RLANG, **kwargs):
        """ Constructor for the class GGplot. """
        data = conversion.py2rpy(data)
        res = cls(cls._constructor(data, mapping=mapping, **kwargs))
        return res

    def plot(self, vp=rpy2.robjects.constants.NULL):
        self._rprint(self, vp=vp)

    def __add__(self, obj):
        res = self._add(self, obj)
        if 'gg' not in res.rclass:
            raise ValueError(
                "Added object did not give a ggplot result "
                "(get class '%s')." % res.rclass[0])
        return self.__class__(res)

    def save(self, filename, **kwargs):
        """ Save the plot ( calls R's `ggplot2::ggsave()` ) """
        ggplot2.ggsave(filename=filename, plot=self, **kwargs)


ggplot = GGPlot.new


class Aes(robjects.ListVector):
    """ Aesthetics mapping, using expressions rather than string
    (this is the most common form when using the package in R - it might
    be easier to use AesString when working in Python using rpy2 -
    see class AesString in this Python module).
    """
    _constructor = ggplot2_env['aes']

    @classmethod
    def new(cls, *args, **kwargs):
        res = cls(cls._constructor(*args, **kwargs))
        return res


aes = Aes.new


class Vars(robjects.ListVector):
    """ Aesthetics mapping, using expressions rather than string
    (this is the most common form when using the package in R - it might
    be easier to use AesString when working in Python using rpy2 -
    see class AesString in this Python module).
    """
    _constructor = ggplot2_env['vars']

    @classmethod
    def new(cls, *args):
        """Constructor for the class Vars."""
        new_args = list()
        for a in args:
            new_args.append(
                rlang.parse_quo(
                    a, env=robjects.baseenv['sys.frame']()
                )
            )
        res = cls(cls._constructor(*new_args))
        return res


vars = Vars.new


class AesString(robjects.ListVector):
    """ Aesthetics mapping, using strings rather than expressions (the later
    being most common form when using the package in R - see class Aes
    in this Python module).

    This associates dimensions in the data sets (columns in the DataFrame),
    possibly with a transformation applied on-the-fly (e.g., "log(value)",
    or "cost / benefits") to graphical "dimensions" in a chosen graphical
    representation (e.g., x-axis, color of points, size, etc...).

    Not all graphical representations have all dimensions. Refer to the
    documentation of ggplot2, online tutorials, or Hadley's book for
    more details.
    """
    _constructor = ggplot2_env['aes_string']

    @classmethod
    def new(cls, **kwargs):
        """Constructor for the class AesString."""
        res = cls(cls._constructor(**kwargs))
        return res


aes_string = AesString.new


class Layer(robjects.RObject):
    """ At this level, aesthetics mapping can (should ?) be specified
     (see Aes and AesString). """
    _constructor = ggplot2_env['layer']

    @classmethod
    def new(cls,
            *args, **kwargs):
        """ Constructor for the class Layer. """
        for i, elt in enumerate(args):
            args[i] = conversion.py2ro(elt)

        for k in kwargs:
            kwargs[k] = conversion.py2ro(kwargs[k])

        res = cls(cls.contructor)(*args, **kwargs)
        return res


layer = Layer.new


class GBaseObject(robjects.Environment):
    @classmethod
    def new(cls, *args, **kwargs):
        args_list = list(args)
        res = cls(cls._constructor(*args_list, **kwargs))
        return res


class Stat(GBaseObject):
    """ A "statistical" processing of the data in order
    to make a plot, or a plot element.

    This is an abstract class; material classes are called
    Stat* (e.g., StatAbline, StatBin, etc...). """
    pass


class StatBin(Stat):
    """ Bin data. """
    _constructor = ggplot2_env['stat_bin']


stat_bin = StatBin.new


class StatBin2D(Stat):
    """ 2D binning of data into squares/rectangles. """
    _constructor = ggplot2_env['stat_bin_2d']


stat_bin2d = StatBin2D.new
stat_bin_2d = StatBin2D.new


class StatBinhex(Stat):
    """ 2D binning of data into hexagons. """
    _constructor = ggplot2_env['stat_bin_hex']


stat_binhex = StatBinhex.new
stat_bin_hex = StatBinhex.new


class StatBoxplot(Stat):
    """ Components of box and whisker plot. """
    _constructor = ggplot2_env['stat_boxplot']


stat_boxplot = StatBoxplot.new


class StatContour(Stat):
    """ Contours of 3D data. """
    _constructor = ggplot2_env['stat_contour']


stat_contour = StatContour.new


class StatDensity(Stat):
    """ 1D density estimate """
    _constructor = ggplot2_env['stat_density']


stat_density = StatDensity.new


class StatDensity2D(Stat):
    """ 2D density estimate """
    _constructor = ggplot2_env['stat_density_2d']


stat_density2d = StatDensity2D.new
stat_density_2d = StatDensity2D.new


class StatFunction(Stat):
    """ Superimpose a function """
    _constructor = ggplot2_env['stat_function']


stat_function = StatFunction.new


class StatIdentity(Stat):
    """ Identity function """
    _constructor = ggplot2_env['stat_identity']


stat_identity = StatIdentity.new


class StatQQ(Stat):
    """ Calculation for quantile-quantile plot. """
    _constructor = ggplot2_env['stat_qq']


stat_qq = StatQQ.new


class StatQuantile(Stat):
    """ Continuous quantiles """
    _constructor = ggplot2_env['stat_quantile']


stat_quantile = StatQuantile.new


class StatSmooth(Stat):
    """ Smoothing function """
    _constructor = ggplot2_env['stat_smooth']


stat_smooth = StatSmooth.new


class StatSpoke(Stat):
    """ Convert angle and radius to xend and yend """
    _constructor = ggplot2_env['stat_spoke']


stat_spoke = StatSpoke.new


class StatSum(Stat):
    """ Sum unique values.
    Useful when overplotting. """
    _constructor = ggplot2_env['stat_sum']


stat_sum = StatSum.new


class StatSummary(Stat):
    """ Summarize values for y at every unique value for x"""
    _constructor = ggplot2_env['stat_summary']


stat_summary = StatSummary.new


class StatSummary2D(Stat):
    """ Summarize values for y at every unique value for x"""
    _constructor = ggplot2_env['stat_summary_2d']


stat_summary2d = StatSummary2D.new
stat_summary_2d = StatSummary2D.new


class StatUnique(Stat):
    """ Remove duplicates. """
    _constructor = ggplot2_env['stat_unique']


stat_unique = StatUnique.new


class Coord(GBaseObject):
    """ Coordinates """
    pass


class CoordFixed(Coord):
    """ Cartesian coordinates with fixed relationship
    (that is fixed ratio between units in axes).
    CoordEquel seems to be identical to this class."""
    _constructor = ggplot2_env['coord_fixed']


coord_fixed = CoordFixed.new


class CoordCartesian(Coord):
    """ Cartesian coordinates. """
    _constructor = ggplot2_env['coord_cartesian']


coord_cartesian = CoordCartesian.new


class CoordEqual(Coord):
    """ This class seems to be identical to CoordFixed. """
    _constructor = ggplot2_env['coord_equal']


coord_equal = CoordEqual.new


class CoordFlip(Coord):
    """ Flip horizontal and vertical coordinates. """
    _constructor = ggplot2_env['coord_flip']


coord_flip = CoordFlip.new


class CoordMap(Coord):
    """ Map projections. """
    _constructor = ggplot2_env['coord_map']


coord_map = CoordMap.new


class CoordPolar(Coord):
    """ Polar coordinates. """
    _constructor = ggplot2_env['coord_polar']


coord_polar = CoordPolar.new


class CoordTrans(Coord):
    """ Apply transformations (functions) to a cartesian coordinate system. """
    _constructor = ggplot2_env['coord_trans']


coord_trans = CoordTrans.new


class Facet(GBaseObject):
    """ Panels """
    pass


class FacetGrid(Facet):
    """ Panels in a grid. """
    _constructor = ggplot2_env['facet_grid']


facet_grid = FacetGrid.new


class FacetWrap(Facet):
    """ Sequence of panels in a 2D layout """
    _constructor = ggplot2_env['facet_wrap']


facet_wrap = FacetWrap.new


class Geom(GBaseObject):
    pass


class GeomAbline(Geom):
    _constructor = ggplot2_env['geom_abline']


geom_abline = GeomAbline.new


class GeomArea(Geom):
    _constructor = ggplot2_env['geom_area']


geom_area = GeomArea.new


class GeomBar(Geom):
    _constructor = ggplot2_env['geom_bar']


geom_bar = GeomBar.new


class GeomCol(Geom):
    _constructor = ggplot2_env['geom_col']


geom_col = GeomCol.new


class GeomBin2D(Geom):
    _constructor = ggplot2_env['geom_bin2d']


geom_bin2d = GeomBin2D.new


class GeomBlank(Geom):
    _constructor = ggplot2_env['geom_blank']


geom_blank = GeomBlank.new


class GeomBoxplot(Geom):
    _constructor = ggplot2_env['geom_boxplot']


geom_boxplot = GeomBoxplot.new


class GeomContour(Geom):
    _constructor = ggplot2_env['geom_contour']


geom_contour = GeomContour.new


class GeomCrossBar(Geom):
    _constructor = ggplot2_env['geom_crossbar']


geom_crossbar = GeomCrossBar.new


class GeomDensity(Geom):
    _constructor = ggplot2_env['geom_density']


geom_density = GeomDensity.new


class GeomDensity2D(Geom):
    _constructor = ggplot2_env['geom_density_2d']


geom_density2d = GeomDensity2D.new
geom_density_2d = GeomDensity2D.new


class GeomDotplot(Geom):
    _constructor = ggplot2_env['geom_dotplot']


geom_dotplot = GeomDotplot.new


class GeomErrorBar(Geom):
    _constructor = ggplot2_env['geom_errorbar']


geom_errorbar = GeomErrorBar.new


class GeomErrorBarH(Geom):
    _constructor = ggplot2_env['geom_errorbarh']


geom_errorbarh = GeomErrorBarH.new


class GeomFreqPoly(Geom):
    _constructor = ggplot2_env['geom_freqpoly']


geom_freqpoly = GeomFreqPoly.new


class GeomHex(Geom):
    _constructor = ggplot2_env['geom_hex']


geom_hex = GeomHex.new


class GeomHistogram(Geom):
    _constructor = ggplot2_env['geom_histogram']


geom_histogram = GeomHistogram.new


class GeomHLine(Geom):
    _constructor = ggplot2_env['geom_hline']


geom_hline = GeomHLine.new


class GeomJitter(Geom):
    _constructor = ggplot2_env['geom_jitter']


geom_jitter = GeomJitter.new


class GeomLine(Geom):
    _constructor = ggplot2_env['geom_line']


geom_line = GeomLine.new


class GeomLineRange(Geom):
    _constructor = ggplot2_env['geom_linerange']


geom_linerange = GeomLineRange.new


class GeomPath(Geom):
    _constructor = ggplot2_env['geom_path']


geom_path = GeomPath.new


class GeomPoint(Geom):
    _constructor = ggplot2_env['geom_point']


geom_point = GeomPoint.new


class GeomPointRange(Geom):
    _constructor = ggplot2_env['geom_pointrange']


geom_pointrange = GeomPointRange.new


class GeomPolygon(Geom):
    _constructor = ggplot2_env['geom_polygon']


geom_polygon = GeomPolygon.new


class GeomQuantile(Geom):
    _constructor = ggplot2_env['geom_quantile']


geom_quantile = GeomQuantile.new


class GeomRaster(Geom):
    _constructor = ggplot2_env['geom_raster']


geom_raster = GeomRaster.new


class GeomRect(Geom):
    _constructor = ggplot2_env['geom_rect']


geom_rect = GeomRect.new


class GeomRibbon(Geom):
    _constructor = ggplot2_env['geom_ribbon']


geom_ribbon = GeomRibbon.new


class GeomRug(Geom):
    _constructor = ggplot2_env['geom_rug']


geom_rug = GeomRug.new


class GeomSegment(Geom):
    _constructor = ggplot2_env['geom_segment']


geom_segment = GeomSegment.new


class GeomSmooth(Geom):
    _constructor = ggplot2_env['geom_smooth']


geom_smooth = GeomSmooth.new


class GeomSpoke(Geom):
    """ Convert angle and radius to xend and yend """
    _constructor = ggplot2_env['geom_spoke']


geom_spoke = GeomSpoke.new


class GeomStep(Geom):
    _constructor = ggplot2_env['geom_step']


geom_step = GeomStep.new


class GeomText(Geom):
    _constructor = ggplot2_env['geom_text']


geom_text = GeomText.new


class GeomTile(Geom):
    _constructor = ggplot2_env['geom_tile']


geom_tile = GeomTile.new


class GeomVLine(Geom):
    _constructor = ggplot2_env['geom_vline']


geom_vline = GeomVLine.new


class Position(GBaseObject):
    pass


class PositionDodge(Position):
    _constructor = ggplot2_env['position_dodge']


position_dodge = PositionDodge.new


class PositionFill(Position):
    _constructor = ggplot2_env['position_fill']


position_fill = PositionFill.new


class PositionJitter(Position):
    _constructor = ggplot2_env['position_jitter']


position_jitter = PositionJitter.new


class PositionStack(Position):
    _constructor = ggplot2_env['position_stack']


position_stack = PositionStack.new


class Scale(GBaseObject):
    pass


class ScaleAlpha(Scale):
    _constructor = ggplot2_env['scale_alpha']


scale_alpha = ScaleAlpha.new


class ScaleColour(Scale):
    pass


class ScaleDiscrete(Scale):
    pass


class ScaleLinetype(Scale):
    _constructor = ggplot2_env['scale_linetype']


scale_linetype = ScaleLinetype.new


class ScaleShape(Scale):
    _constructor = ggplot2_env['scale_shape']


scale_shape = ScaleShape.new


class ScaleRadius(Scale):
    _constructor = ggplot2_env['scale_radius']


scale_radius = ScaleRadius.new


class ScaleSize(Scale):
    _constructor = ggplot2_env['scale_size']


scale_size = ScaleSize.new


class ScaleSizeArea(Scale):
    _constructor = ggplot2_env['scale_size_area']


scale_size_area = ScaleSizeArea.new


class ScaleShapeDiscrete(Scale):
    _constructor = ggplot2_env['scale_shape_discrete']


scale_shape_discrete = ScaleShapeDiscrete.new


class ScaleFill(Scale):
    pass


class ScaleX(Scale):
    pass


class ScaleY(Scale):
    pass

# class Limits(Scale):
#    _constructor = ggplot2_env['limits']
# limits = Limits.new


class XLim(Scale):
    _constructor = ggplot2_env['xlim']


xlim = XLim.new


class YLim(Scale):
    _constructor = ggplot2_env['ylim']


ylim = YLim.new


class ScaleXContinuous(ScaleX):
    _constructor = ggplot2_env['scale_x_continuous']


scale_x_continuous = ScaleXContinuous.new


class ScaleYContinuous(ScaleY):
    _constructor = ggplot2_env['scale_y_continuous']


scale_y_continuous = ScaleYContinuous.new


class ScaleXDiscrete(ScaleX):
    _constructor = ggplot2_env['scale_x_discrete']


scale_x_discrete = ScaleXDiscrete.new


class ScaleYDiscrete(ScaleY):
    _constructor = ggplot2_env['scale_y_discrete']


scale_y_discrete = ScaleYDiscrete.new


class ScaleXDate(ScaleX):
    _constructor = ggplot2_env['scale_x_date']


scale_x_date = ScaleXDate.new


class ScaleYDate(ScaleY):
    _constructor = ggplot2_env['scale_y_date']


scale_y_date = ScaleYDate.new


class ScaleXDatetime(ScaleX):
    _constructor = ggplot2_env['scale_x_datetime']


scale_x_datetime = ScaleXDatetime.new


class ScaleYDatetime(ScaleY):
    _constructor = ggplot2_env['scale_y_datetime']


scale_y_datetime = ScaleYDatetime.new


class ScaleXLog10(ScaleX):
    _constructor = ggplot2_env['scale_x_log10']


scale_x_log10 = ScaleXLog10.new


class ScaleYLog10(ScaleY):
    _constructor = ggplot2_env['scale_y_log10']


scale_y_log10 = ScaleYLog10.new


class ScaleXReverse(ScaleX):
    _constructor = ggplot2_env['scale_x_reverse']


scale_x_reverse = ScaleXReverse.new


class ScaleYReverse(ScaleY):
    _constructor = ggplot2_env['scale_y_reverse']


scale_y_reverse = ScaleYReverse.new


class ScaleXSqrt(ScaleX):
    _constructor = ggplot2_env['scale_x_sqrt']


scale_x_sqrt = ScaleXSqrt.new


class ScaleYSqrt(ScaleY):
    _constructor = ggplot2_env['scale_y_sqrt']


scale_y_sqrt = ScaleYSqrt.new


class ScaleColourBrewer(ScaleColour):
    _constructor = ggplot2_env['scale_colour_brewer']


scale_colour_brewer = ScaleColourBrewer.new
scale_color_brewer = scale_colour_brewer


class ScaleColourContinuous(ScaleColour):
    _constructor = ggplot2_env['scale_colour_continuous']


scale_colour_continuous = ScaleColourContinuous.new
scale_color_continuous = scale_colour_continuous


class ScaleColourDiscrete(ScaleColour):
    _constructor = ggplot2_env['scale_colour_discrete']


scale_colour_discrete = ScaleColourDiscrete.new
scale_color_discrete = scale_colour_discrete


class ScaleColourGradient(ScaleColour):
    _constructor = ggplot2_env['scale_colour_gradient']


scale_colour_gradient = ScaleColourGradient.new
scale_color_gradient = scale_colour_gradient


class ScaleColourGradient2(ScaleColour):
    _constructor = ggplot2_env['scale_colour_gradient2']


scale_colour_gradient2 = ScaleColourGradient2.new
scale_color_gradient2 = scale_colour_gradient2


class ScaleColourGradientN(ScaleColour):
    _constructor = ggplot2_env['scale_colour_gradientn']


scale_colour_gradientn = ScaleColourGradientN.new
scale_color_gradientn = scale_colour_gradientn


class ScaleColourGrey(ScaleColour):
    _constructor = ggplot2_env['scale_colour_grey']


scale_colour_grey = ScaleColourGrey.new
scale_color_grey = scale_colour_grey


class ScaleColourHue(ScaleColour):
    _constructor = ggplot2_env['scale_colour_hue']


scale_colour_hue = ScaleColourHue.new
scale_color_hue = scale_colour_hue


class ScaleColourIdentity(ScaleColour):
    _constructor = ggplot2_env['scale_colour_identity']


scale_colour_identity = ScaleColourIdentity.new
scale_color_identity = scale_colour_identity


class ScaleColourManual(ScaleColour):
    _constructor = ggplot2_env['scale_colour_manual']


scale_colour_manual = ScaleColourManual.new
scale_color_manual = scale_colour_manual


class ScaleFillBrewer(ScaleFill):
    _constructor = ggplot2_env['scale_fill_brewer']


scale_fill_brewer = ScaleFillBrewer.new


class ScaleFillContinuous(ScaleFill):
    _constructor = ggplot2_env['scale_fill_continuous']


scale_fill_continuous = ScaleFillContinuous.new


class ScaleFillDiscrete(ScaleFill):
    _constructor = ggplot2_env['scale_fill_discrete']


scale_fill_discrete = ScaleFillDiscrete.new


class ScaleFillGradient(ScaleFill):
    _constructor = ggplot2_env['scale_fill_gradient']


scale_fill_gradient = ScaleFillGradient.new


class ScaleFillGradient2(ScaleFill):
    _constructor = ggplot2_env['scale_fill_gradient2']


scale_fill_gradient2 = ScaleFillGradient2.new


class ScaleFillGradientN(ScaleFill):
    _constructor = ggplot2_env['scale_fill_gradientn']


scale_fill_gradientn = ScaleFillGradientN.new


class ScaleFillGrey(ScaleFill):
    _constructor = ggplot2_env['scale_fill_grey']


scale_fill_grey = ScaleFillGrey.new


class ScaleFillHue(ScaleFill):
    _constructor = ggplot2_env['scale_fill_hue']


scale_fill_hue = ScaleFillHue.new


class ScaleFillIdentity(ScaleFill):
    _constructor = ggplot2_env['scale_fill_identity']


scale_fill_identity = ScaleFillIdentity.new


class ScaleFillManual(ScaleFill):
    _constructor = ggplot2_env['scale_fill_manual']


scale_fill_manual = ScaleFillManual.new


class ScaleLinetypeContinuous(ScaleLinetype):
    _constructor = ggplot2_env['scale_linetype_continuous']


scale_linetype_continuous = ScaleLinetypeContinuous.new


class ScaleLinetypeDiscrete(ScaleLinetype):
    _constructor = ggplot2_env['scale_linetype_discrete']


scale_linetype_discrete = ScaleLinetypeDiscrete.new


class ScaleLinetypeManual(ScaleLinetype):
    _constructor = ggplot2_env['scale_linetype_manual']


scale_linetype_manual = ScaleLinetypeManual.new


class ScaleShapeManual(ScaleShape):
    _constructor = ggplot2_env['scale_shape_manual']


scale_shape_manual = ScaleShapeManual.new


guides = ggplot2.guides
guide_colorbar = ggplot2.guide_colorbar
guide_colourbar = ggplot2.guide_colourbar
guide_legend = ggplot2.guide_legend


class Options(robjects.ListVector):

    def __repr__(self):
        s = '<instance of %s : %i>' % (type(self), id(self))
        return s


class Element(Options):
    pass


class ElementText(Element):

    _constructor = ggplot2.element_text

    @classmethod
    def new(cls, family='', face='plain', colour='black', size=10,
            hjust=0.5, vjust=0.5, angle=0, lineheight=1.1,
            color=NULL):
        res = cls(cls._constructor(family=family, face=face,
                                   colour=colour, size=size,
                                   hjust=hjust, vjust=vjust,
                                   angle=angle, lineheight=lineheight))
        return res


element_text = ElementText.new


class ElementRect(Element):

    _constructor = ggplot2.element_rect

    @classmethod
    def new(cls, fill=NULL, colour=NULL, size=NULL,
            linetype=NULL, color=NULL):
        res = cls(cls._constructor(fill=fill, colour=colour,
                                   size=size, linetype=linetype,
                                   color=color))
        return res


element_rect = ElementRect.new


class Labs(Options):
    _constructor = ggplot2.labs

    @classmethod
    def new(cls, **kwargs):
        res = cls(cls._constructor(**kwargs))
        return res


labs = Labs.new


class Theme(Options):
    pass


class ElementBlank(Theme):
    _constructor = ggplot2.element_blank

    @classmethod
    def new(cls):
        res = cls(cls._constructor())
        return res


element_blank = ElementBlank.new


theme_get = ggplot2.theme_get


class ThemeGrey(Theme):
    _constructor = ggplot2.theme_grey

    @classmethod
    def new(cls, base_size=12):
        res = cls(cls._constructor(base_size=base_size))
        return res


theme_grey = ThemeGrey.new


class ThemeClassic(Theme):
    _constructor = ggplot2.theme_classic

    @classmethod
    def new(cls, base_size=12, base_family=''):
        res = cls(cls._constructor(base_size=base_size,
                                   base_family=base_family))
        return res


theme_classic = ThemeClassic.new


class ThemeDark(Theme):
    _constructor = ggplot2.theme_dark

    @classmethod
    def new(cls, base_size=12, base_family=''):
        res = cls(cls._constructor(base_size=base_size,
                                   base_family=base_family))
        return res


theme_dark = ThemeDark.new


class ThemeLight(Theme):
    _constructor = ggplot2.theme_light

    @classmethod
    def new(cls, base_size=12, base_family=''):
        res = cls(cls._constructor(base_size=base_size,
                                   base_family=base_family))
        return res


theme_light = ThemeLight.new


class ThemeBW(Theme):
    _constructor = ggplot2.theme_bw

    @classmethod
    def new(cls, base_size=12):
        res = cls(cls._constructor(base_size=base_size))
        return res


theme_bw = ThemeBW.new


class ThemeGray(Theme):
    _constructor = ggplot2.theme_gray

    @classmethod
    def new(cls, base_size=12):
        res = cls(cls._constructor(base_size=base_size))
        return res


theme_gray = ThemeGray.new
theme_grey = theme_gray


class ThemeMinimal(Theme):
    _constructor = ggplot2.theme_minimal

    @classmethod
    def new(cls, base_size=12, base_family=''):
        res = cls(cls._constructor(base_size=base_size,
                                   base_family=base_family))
        return res


theme_minimal = ThemeMinimal.new


class ThemeLinedraw(Theme):
    _constructor = ggplot2.theme_linedraw

    @classmethod
    def new(cls, base_size=12, base_family=""):
        res = cls(cls._constructor(base_size=base_size,
                                   base_family=base_family))
        return res


theme_linedraw = ThemeLinedraw.new


class ThemeVoid(Theme):
    _constructor = ggplot2.theme_void

    @classmethod
    def new(cls, base_size=12, base_family=''):
        res = cls(cls._constructor(base_size=base_size,
                                   base_family=base_family))
        return res


theme_void = ThemeVoid.new


theme_set = ggplot2.theme_set

theme_update = ggplot2.theme_update

gplot = ggplot2.qplot

map_data = ggplot2.map_data

theme = ggplot2_env['theme']

ggtitle = ggplot2.ggtitle
xlab = ggplot2.xlab
ylab = ggplot2.ylab
guide_axis = ggplot2.guide_axis
guide_bins = ggplot2.guide_bins
guide_colorbar = ggplot2.guide_colorbar
guide_colourbar = guide_colorbar
guide_colorsteps = ggplot2.guide_colorsteps
guide_coloursteps = guide_colorsteps
guide_geom = ggplot2.guide_geom
guide_legend = ggplot2.guide_legend
guide_merge = ggplot2.guide_merge
guide_none = ggplot2.guide_none
guide_train = ggplot2.guide_train
guide_transform = ggplot2.guide_transform


def dict2rvec(d):
    """Convert a python dict[str, str] into an R named vector.
    """
    r_x = StrVector(list(d.values()))
    r_x.names = list(d.keys())
    return r_x


as_labeller = ggplot2.as_labeller

original_rpy2py = conversion.rpy2py


def ggplot2_conversion(robj):

    pyobj = original_rpy2py(robj)

    try:
        rcls = pyobj.rclass
    except AttributeError:
        # conversion lead to something that is no
        # longer an R object
        return pyobj

    if (rcls is not None) and (rcls[0] == 'gg'):
        pyobj = GGPlot(pyobj)

    return pyobj


conversion.rpy2py = ggplot2_conversion



""" Utility module with functions related to R packages
(having these in this utility module rather than in packages.py
prevents circular imports). """

from rpy2 import rinterface
from warnings import warn
from collections import defaultdict

_packages = rinterface.baseenv['.packages']
_libpaths = rinterface.baseenv['.libPaths']
_find_package = rinterface.baseenv['find.package']


def get_packagepath(package: str) -> str:
    """ return the path to an R package installed """
    res = _find_package(rinterface.StrSexpVector((package, )))
    return res[0]


# Functions to translate R symbols to Python symbols.
# The functions are in this module in order to facilitate
# their access from other modules (without circular dependencies).
# It is not necessarily the absolute best place to have the functions though.
def default_symbol_r2python(rname: str) -> str:
    """Replace each dot (.) with an underscore (_)."""
    return rname.replace('.', '_')


def default_symbol_resolve(symbol_mapping):
    """Resolve any conflict in a symbol mapping.

    The argument `symbol_mapping` maps candidate new symbol names
    (e.g., the names of Python attributes in the namespace returned by
    :func:`importr`)
    to a sequence of original symbol names (e.g., the names of objects in
    an R package). The purpose of this function is to resolved conflicts,
    that is situations where there is more than one original symbol name
    associated with a new symbol name.

    :param symbol_mapping: a :class:`dict` or dict-like object.
    :return: A 2-tuple with conflicts (a :class:`dict` mapping the new
    symbol to a sequence of symbols already matching) and resolutions (a
    :class:`dict` mapping new symbol).
    """
    # dict to store the Python symbol -> R symbols mapping causing problems.
    conflicts = dict()
    resolutions = dict()
    for py_symbol, r_symbols in symbol_mapping.items():
        n_r_symbols = len(r_symbols)
        if n_r_symbols == 1:
            continue
        elif n_r_symbols == 2:
            # more than one R symbol associated with this Python symbol
            try:
                idx = r_symbols.index(py_symbol)
                # there is an R symbol identical to the proposed Python symbol;
                # we keep that pair mapped, and change the Python symbol for
                # the other R symbol(s) according to PEP 0008
                for i, r_name in enumerate(r_symbols):
                    if i == idx:
                        resolutions[py_symbol] = [r_name, ]
                    else:
                        new_py_symbol = py_symbol + '_'
                        resolutions[new_py_symbol] = [r_name, ]
            except ValueError:
                # I am unsure about what to do at this point:
                # add it as a conflict
                conflicts[py_symbol] = r_symbols
        else:
            # no automatic resolution if more than 2
            conflicts[py_symbol] = r_symbols
    return conflicts, resolutions


def _map_symbols(rnames,
                 translation=dict(),
                 symbol_r2python=default_symbol_r2python,
                 symbol_resolve=default_symbol_resolve):
    """
    :param names: an iterable of rnames
    :param translation: a mapping for R name->python name
    :param symbol_r2python: a function to translate an R symbol into a
                            (presumably valid) Python symbol
    :param symbol_resolve: a function to check a prospective set of
                           translation and resolve conflicts if needed
    """
    symbol_mapping = defaultdict(list)
    for rname in rnames:
        if rname in translation:
            rpyname = translation[rname]
        else:
            rpyname = symbol_r2python(rname)
        symbol_mapping[rpyname].append(rname)
    conflicts, resolutions = symbol_resolve(symbol_mapping)

    return (symbol_mapping, conflicts, resolutions)


def _fix_map_symbols(symbol_mapping,
                     conflicts,
                     on_conflict,
                     msg_prefix,
                     exception):
    """
    :param symbol_mapping: as returned by `_map_symbols`
    :param conflicts: as returned by `_map_symbols`
    :param on_conflict: action to take if conflict
    :param msg_prefix: prefix for error message
    :param exception: exception to raise
    """
    if len(conflicts) > 0:
        msg = msg_prefix
        msg += '\n- '.join(('%s -> %s' %
                            (k, ', '.join(v))
                            for k, v in conflicts.items()))
        if on_conflict == 'fail':
            msg += ('\nTo turn this exception into a simple '
                    'warning use the parameter '
                    '`on_conflict="warn"`')
            raise exception(msg)
        elif on_conflict == 'warn':
            for k, v in conflicts.items():
                if k in v:
                    symbol_mapping[k] = [k, ]
                else:
                    del(symbol_mapping[k])
            warn(msg)
        else:
            raise ValueError('Invalid value for parameter "on_conflict"')



import pytest
import sys
from rpy2 import robjects
from rpy2 import rinterface
import rpy2.robjects.conversion as conversion
r = robjects.r


class DummyNamespace(object):
    def __getattr__(self, name):
        return None


has_numpy = False
try:
    import numpy
    has_numpy = True
    import rpy2.robjects.numpy2ri as rpyn
except:
    numpy = DummyNamespace()


@pytest.fixture()
def numpy_conversion():
    with conversion.localconverter(
            robjects.default_converter + rpyn.converter
    ) as lc:
        yield


@pytest.mark.skipif(not has_numpy,
                    reason='package numpy cannot be imported')
@pytest.mark.usefixtures('numpy_conversion')
class TestNumpyConversions(object):

    def check_homogeneous(self, obj, mode, storage_mode):
        converted = conversion.py2rpy(obj)
        assert r["mode"](converted)[0] == mode
        assert r["storage.mode"](converted)[0] == storage_mode
        assert list(obj) == list(converted)
        assert r["is.array"](converted)[0] is True
        return converted

    def test_vector_boolean(self):
        l = [True, False, True]
        b = numpy.array(l, dtype=numpy.bool_)
        b_r = self.check_homogeneous(b, "logical", "logical")
        assert tuple(l) == tuple(b_r)

    def test_vector_integer(self):
        l = [1, 2, 3]
        i = numpy.array(l, dtype="i")
        i_r = self.check_homogeneous(i, "numeric", "integer")
        assert tuple(l) == tuple(i_r)
        
    def test_vector_float(self):
        l = [1.0, 2.0, 3.0]
        f = numpy.array(l, dtype="f")
        f_r = self.check_homogeneous(f, "numeric", "double")
        for orig, conv in zip(l, f_r):
            assert abs(orig-conv) < 0.000001
        
    def test_vector_complex(self):
        l = [1j, 2j, 3j]
        c = numpy.array(l, dtype=numpy.complex_)
        c_r = self.check_homogeneous(c, "complex", "complex")
        for orig, conv in zip(l, c_r):
            assert abs(orig.real-conv.real) < 0.000001
            assert abs(orig.imag-conv.imag) < 0.000001

    def test_vector_unicode_character(self):
        l = [u"a", u"c", u"e"]
        u = numpy.array(l, dtype="U")
        u_r = self.check_homogeneous(u, "character", "character")
        assert tuple(l) == tuple(u_r)

    def test_vector_bytes(self):
        l = [b'a', b'b', b'c']
        s = numpy.array(l, dtype = '|S1')
        converted = conversion.py2rpy(s)
        assert r["mode"](converted)[0] == 'raw'
        assert r["storage.mode"](converted)[0] == 'raw'
        assert bytearray(b''.join(l)) == bytearray(converted)

    def test_array(self):

        i2d = numpy.array([[1, 2, 3], [4, 5, 6]], dtype='i')
        i2d_r = conversion.py2rpy(i2d)
        assert r['storage.mode'](i2d_r)[0] == 'integer'
        assert tuple(r['dim'](i2d_r)) == (2, 3)

        # Make sure we got the row/column swap right:
        assert r['['](i2d_r, 1, 2)[0] == i2d[0, 1]

        f3d = numpy.arange(24, dtype='f').reshape((2, 3, 4))
        f3d_r = conversion.py2rpy(f3d)

        assert r['storage.mode'](f3d_r)[0] == 'double'
        assert tuple(r['dim'](f3d_r)) == (2, 3, 4)

        # Make sure we got the row/column swap right:
        #assert r['['](f3d_r, 1, 2, 3)[0] == f3d[0, 1, 2]

    @pytest.mark.skipif(not has_numpy,
                        reason='package numpy cannot be imported')
    @pytest.mark.parametrize(
        'constructor',
        (numpy.int32, numpy.int64,
         numpy.uint32, numpy.uint64)
    )
    def test_scalar_int(self, constructor):
        np_value = constructor(100)
        r_vec = conversion.py2rpy(np_value)
        r_scalar = numpy.array(r_vec)[0]
        assert np_value == r_scalar

    @pytest.mark.skipif(not (has_numpy and hasattr(numpy, 'float128')),
                        reason='numpy.float128 not available on this system')
    def test_scalar_f128(self):
        f128 = numpy.float128(100.000000003)
        f128_r = conversion.py2rpy(f128)
        f128_test = numpy.array(f128_r)[0]
        assert f128 == f128_test

    def test_object_array(self):
        o = numpy.array([1, "a", 3.2], dtype=numpy.object_)
        o_r = conversion.py2rpy(o)
        assert r['mode'](o_r)[0] == 'list'
        assert r['[['](o_r, 1)[0] == 1
        assert r['[['](o_r, 2)[0] == 'a'
        assert r['[['](o_r, 3)[0] == 3.2

    def test_record_array(self):
        rec = numpy.array([(1, 2.3), (2, -0.7), (3, 12.1)],
                          dtype=[("count", "i"), ("value", numpy.double)])
        rec_r = conversion.py2rpy(rec)
        assert r["is.data.frame"](rec_r)[0] is True
        assert tuple(r["names"](rec_r)) == ("count", "value")
        count_r = rec_r[rec_r.names.index('count')]
        value_r = rec_r[rec_r.names.index('value')]
        assert r["storage.mode"](count_r)[0] == 'integer'
        assert r["storage.mode"](value_r)[0] == 'double'
        assert count_r[1] == 2
        assert value_r[2] == 12.1

    def test_bad_array(self):
        u = numpy.array([1, 2, 3], dtype=numpy.uint32)
        with pytest.raises(ValueError):
            conversion.py2rpy(u)

    def test_assign_numpy_object(self):
        x = numpy.arange(-10., 10., 1)
        env = robjects.Environment()
        env['x'] = x
        assert len(env) == 1
        # do have an R object of the right type ?
        with conversion.localconverter(
            robjects.default_converter
        ) as lc:
            x_r = env['x']

        assert robjects.rinterface.RTYPES.REALSXP == x_r.typeof
        #
        assert tuple(x_r.dim) == (20,)


    def test_dataframe_to_numpy(self):
        df = robjects.vectors.DataFrame(
            {'a': 1,
             'b': 2,
             'c': robjects.vectors.FactorVector('e')})
        rec = conversion.rpy2py(df)
        assert numpy.recarray == type(rec)
        assert rec.a[0] == 1
        assert rec.b[0] == 2
        assert rec.c[0] == 'e'

    def test_atomic_vector_to_numpy(self):
        v = robjects.vectors.IntVector((1,2,3))
        a = rpyn.rpy2py(v)
        assert isinstance(a, numpy.ndarray)
        assert v[0] == 1

    def test_rx2(self):
        df = robjects.vectors.DataFrame({
            "A": robjects.vectors.IntVector([1,2,3]),
            "B": robjects.vectors.IntVector([1,2,3])})
        b = df.rx2('B')
        assert tuple((1,2,3)) == tuple(b)

    def test_rint_to_numpy(self):
        a = robjects.r('c(1:4)')
        assert isinstance(a, numpy.ndarray)

    def test_rfloat_to_numpy(self):
        a = robjects.r('c(1.0, 2.0, 3.0)')
        assert isinstance(a, numpy.ndarray)


@pytest.mark.skipif(not has_numpy,
                    reason='package numpy cannot be imported')
@pytest.mark.parametrize('dtype',
                         ('uint32', 'uint64'))
def test_unsignednumpyint_to_rint_error(dtype):
    values = (1,2,3)
    a = numpy.array(values, dtype=dtype)
    with pytest.raises(ValueError):
        rpyn.unsignednumpyint_to_rint(a)


@pytest.mark.skipif(not has_numpy,
                    reason='package numpy cannot be imported')
@pytest.mark.parametrize('dtype',
                         ('uint8', 'uint16'))
def test_unsignednumpyint_to_rint(dtype):
    values = (1,2,3)
    a = numpy.array(values, dtype=dtype)
    v = rpyn.unsignednumpyint_to_rint(a)
    assert values == tuple(v)


@pytest.mark.skipif(not has_numpy,
                    reason='package numpy cannot be imported')
@pytest.mark.parametrize('values,expected_cls',
                         ((['a', 1, 2], robjects.vectors.ListVector),
                          (['a', 'b', 'c'], rinterface.StrSexpVector),
                          ([b'a', b'b', b'c'], rinterface.ByteSexpVector)))
def test_numpy_O_py2rpy(values, expected_cls):
    a = numpy.array(values, dtype='O')
    v = rpyn.numpy_O_py2rpy(a)
    assert isinstance(v, expected_cls)

    



import rpy2.rinterface as ri


ri.initr()


def test_init_from_r():
    pairlist = ri.baseenv.find('pairlist')
    pl = pairlist(a=ri.StrSexpVector(['1', ]),
                  b=ri.StrSexpVector(['3', ]))
    assert pl.typeof == ri.RTYPES.LISTSXP


def test_names():
    pairlist = ri.baseenv.find('pairlist')
    pl = pairlist(a=ri.StrSexpVector(['1', ]),
                  b=ri.StrSexpVector(['3', ]))
    assert tuple(pl.names) == ('a', 'b')


def test_getitem_pairlist():
    pairlist = ri.baseenv.find('pairlist')
    pl = pairlist(a=ri.StrSexpVector(['1', ]),
                  b=ri.StrSexpVector(['3', ]))
    # R's behaviour is that subsetting returns an R list
    y = pl[0]
    assert y.typeof == ri.RTYPES.VECSXP
    assert len(y) == 1
    assert y[0][0] == '1'
    assert y.names[0] == 'a'


def test_getslice_pairlist():
    pairlist = ri.baseenv.find('pairlist')
    vec = pairlist(a=ri.StrSexpVector(['1', ]),
                   b=ri.StrSexpVector(['3', ]),
                   c=ri.StrSexpVector(['6', ]))
    vec_slice = vec[0:2]
    assert vec_slice.typeof == ri.RTYPES.LISTSXP
    assert len(vec_slice) == 2
    assert tuple(vec_slice[0][0]) == ('1', )
    assert tuple(vec_slice[1][0]) == ('3', )
    assert tuple(vec_slice.names) == ('a', 'b')


def test_getslice_pairlist_issue380():
    # Checks that root of issue #380 is fixed
    vec = ri.baseenv['.Options']
    vec_slice = vec[0:2]
    assert len(vec_slice) == 2
    assert vec_slice.typeof == ri.RTYPES.LISTSXP
    assert vec.names[0] == vec_slice.names[0]
    assert vec.names[1] == vec_slice.names[1]



import array
import pytest
import rpy2.rinterface_lib.sexp
from rpy2 import rinterface
from rpy2 import robjects


@pytest.mark.parametrize(
    'cls, values, expected_cls',
    [(rinterface.IntSexpVector, (1, 2), robjects.vectors.IntVector),
     (rinterface.FloatSexpVector, (1.1, 2.2), robjects.vectors.FloatVector),
     (rinterface.StrSexpVector, ('ab', 'cd'), robjects.vectors.StrVector),
     (rinterface.BoolSexpVector, (True, False), robjects.vectors.BoolVector),
     (rinterface.ByteSexpVector, b'ab', robjects.vectors.ByteVector),
     (lambda x: rinterface.evalr(x), 'y ~ x', robjects.Formula)])
def test_sexpvector_to_ro(cls, values, expected_cls):
    v_ri = cls(values)
    v_ro = robjects.default_converter.rpy2py(v_ri)
    assert isinstance(v_ro, expected_cls)


def test_mapperR2Python_function():
    sexp = rinterface.globalenv.find('plot')
    ob = robjects.default_converter.rpy2py(sexp)
    assert isinstance(ob, robjects.Function)

    
def test_mapperR2Python_environment():
    sexp = rinterface.globalenv.find('.GlobalEnv')
    assert isinstance(robjects.default_converter.rpy2py(sexp), 
                      robjects.Environment)


def test_mapperR2Python_lang():
    sexp = rinterface.baseenv['str2lang']('1+2')
    ob = robjects.default_converter.rpy2py(sexp)
    assert isinstance(ob, robjects.language.LangVector)


def test_mapperR2Python_Date():
    sexp = rinterface.baseenv.find('as.Date')('2020-01-01')
    assert isinstance(robjects.default_converter.rpy2py(sexp), 
                      robjects.vectors.DateVector)


def test_NameClassMap():
    ncm = robjects.conversion.NameClassMap(object)
    classnames = ('A', 'B')
    assert ncm.find_key(classnames) is None
    assert ncm.find(classnames) is object
    ncm['B'] = list
    assert ncm.find_key(classnames) == 'B'
    assert ncm.find(classnames) is list
    ncm['A'] = tuple
    assert ncm.find_key(classnames) == 'A'
    assert ncm.find(classnames) is tuple


def test_NameClassMapContext():
    ncm = robjects.conversion.NameClassMap(object)
    assert not len(ncm._map)
    with robjects.conversion.NameClassMapContext(ncm, {}):
        assert not len(ncm._map)
    assert not len(ncm._map)
    
    with robjects.conversion.NameClassMapContext(ncm, {'A': list}):
        assert set(ncm._map.keys()) == set('A')
    assert not len(ncm._map)
    ncm['B'] = tuple
    with robjects.conversion.NameClassMapContext(ncm, {'A': list}):
        assert set(ncm._map.keys()) == set('AB')
    assert set(ncm._map.keys()) == set('B')
    with robjects.conversion.NameClassMapContext(ncm, {'B': list}):
        assert set(ncm._map.keys()) == set('B')
    assert set(ncm._map.keys()) == set('B')
    assert ncm['B'] is tuple


def test_Converter_rclass_map_context():
    converter = robjects.default_converter

    class FooEnv(robjects.Environment):
        pass

    ncm = converter.rpy2py_nc_name[rinterface.SexpEnvironment]
    with robjects.default_converter.rclass_map_context(
            rinterface.SexpEnvironment, {'A': FooEnv}
    ):
        assert set(ncm._map.keys()) == set('A')
    assert not len(ncm._map)


@pytest.fixture(scope='module')
def _set_class_AB():
    robjects.r('A <- methods::setClass("A", representation(x="integer"))')
    robjects.r('B <- methods::setClass("B", contains="A")')
    yield
    robjects.r('methods::removeClass("B")')
    robjects.r('methods::removeClass("A")')


def test_mapperR2Python_s4(_set_class_AB):
    classname = rinterface.StrSexpVector(['A', ])
    one = rinterface.IntSexpVector([1, ])
    sexp = rinterface.globalenv['A'](x=one)
    assert isinstance(robjects.default_converter.rpy2py(sexp), 
                      robjects.RS4)


def test_mapperR2Python_s4custom(_set_class_AB):
    class A(robjects.RS4):
        pass
    sexp_a = rinterface.globalenv['A']( 
        x=rinterface.IntSexpVector([1, ])
    )
    sexp_b = rinterface.globalenv['B']( 
        x=rinterface.IntSexpVector([2, ])
    )
    rs4_map = robjects.conversion.converter.rpy2py_nc_name[rinterface.SexpS4]
    with robjects.conversion.NameClassMapContext(
            rs4_map,
            {'A': A}
    ):
        assert rs4_map.find_key(('A', )) == 'A'
        assert isinstance(
            robjects.default_converter.rpy2py(sexp_a), 
            A)
        assert rs4_map.find_key(('B', 'A')) == 'A'
        assert isinstance(
            robjects.default_converter.rpy2py(sexp_b), 
            A)

@pytest.mark.parametrize('value,cls', [
    (1, int),
    (True, bool),
    (b'houba', bytes),
    (1.0, float),
    (1.0 + 2j, complex)
])
def test_py2ro_mappedtype(value, cls):
    pyobj = value
    assert isinstance(pyobj, cls)
    rob = robjects.default_converter.py2rpy(pyobj)
    assert isinstance(rob, cls)
                    

def test_mapperPy2R_taggedlist():
    py = robjects.rlc.TaggedList(('a', 'b'),
                                 tags=('foo', 'bar'))
    robj = robjects.default_converter.py2rpy(py)
    assert isinstance(robj, robjects.Vector)
    assert len(robj) == 2
    assert tuple(robj.names) == ('foo', 'bar')


def test_mapperPy2R_function():
    func = lambda x: x
    rob = robjects.default_converter.py2rpy(func)
    assert isinstance(rob, robjects.SignatureTranslatedFunction)
    assert rob.typeof == rinterface.RTYPES.CLOSXP


@pytest.mark.parametrize('ctype', ['h', 'H', 'i', 'I'])
def test_mapperpy2rpy_int_array(ctype):
    a = array.array(ctype, range(10))
    rob = robjects.default_converter.py2rpy(a)
    assert isinstance(rob, robjects.vectors.IntSexpVector)
    assert isinstance(rob, robjects.vectors.IntVector)
    assert rob.typeof == rinterface.RTYPES.INTSXP


@pytest.mark.parametrize('ctype', ['d', 'f'])
def test_mapperpy2rpy_float_array(ctype):
    a = array.array(ctype, (1.1, 2.2, 3.3))
    rob = robjects.default_converter.py2rpy(a)
    assert isinstance(rob, robjects.vectors.FloatSexpVector)
    assert isinstance(rob, robjects.vectors.FloatVector)
    assert rob.typeof == rinterface.RTYPES.REALSXP

    
def noconversion():
    robj_res = robjects.baseenv['pi']
    assert isinstance(robj_res, robjects.RObject)
    rint_res = robject.conversion.noconversion(robj_res)
    assert isinstance(rint_res, rpy2.rinterface_lib.sexp.Sexp)



# TODO: make it cffi-buildable with conditional function definition
# (Python if ABI, C if API)
import abc
import enum
import logging
from typing import Tuple
import typing
import warnings
from rpy2.rinterface_lib import ffi_proxy
from rpy2.rinterface_lib import openrlib
from rpy2.rinterface_lib import conversion
from rpy2.rinterface_lib import embedded
from rpy2.rinterface_lib import memorymanagement

from _cffi_backend import FFI  # type: ignore


logger = logging.getLogger(__name__)

ffi = openrlib.ffi

# TODO: How can I reliably get MAX_INT from limits.h ?
_MAX_INT = 2**32-1

_R_PRESERVED = dict()  # type: typing.Dict[int, int]
_PY_PASSENGER = dict()

FFI_MODE = ffi_proxy.get_ffi_mode(openrlib._rinterface_cffi)


def get_rid(cdata: FFI.CData) -> int:
    """Get the identifier for the R object.

    This is intended to be like Python's `id()`, but
    for R objects mapped by rpy2."""
    return int(ffi.cast('uintptr_t', cdata))


def protected_rids() -> Tuple[Tuple[int, int], ...]:
    """Sequence of R IDs protected from collection by rpy2."""
    keys = tuple(_R_PRESERVED.keys())
    res = []
    for k in keys:
        v = _R_PRESERVED.get(k)
        if v:
            res.append((get_rid(k), v))
    return tuple(res)


def is_cdata_sexp(obj: typing.Any) -> bool:
    """Is the object a cffi `CData` object pointing to an R object ?"""
    if (isinstance(obj, FFI.CData) and
            ffi.typeof(obj).cname == 'struct SEXPREC *'):
        return True
    else:
        return False


def _preserve(cdata: FFI.CData) -> int:
    addr = int(ffi.cast('uintptr_t', cdata))
    count = _R_PRESERVED.get(addr, 0)
    if count == 0:
        openrlib.rlib.R_PreserveObject(cdata)
    _R_PRESERVED[addr] = count + 1
    return addr


def _release(cdata: FFI.CData) -> None:
    addr = int(ffi.cast('uintptr_t', cdata))
    count = _R_PRESERVED[addr] - 1
    if count == 0:
        del(_R_PRESERVED[addr])
        openrlib.rlib.R_ReleaseObject(cdata)
    else:
        _R_PRESERVED[addr] = count


@ffi_proxy.callback(ffi_proxy._capsule_finalizer_def,
                    openrlib._rinterface_cffi)
def _capsule_finalizer(cdata: FFI.CData) -> None:
    try:
        openrlib.rlib.R_ClearExternalPtr(cdata)
    except Exception as e:
        warnings.warn('Exception downgraded to warning: %s' % str(e))


# ABI and API modes differs in what is the exact callback object to be
# passed to C code.
if hasattr(openrlib._rinterface_cffi, 'lib'):
    _capsule_finalizer_c = openrlib._rinterface_cffi.lib._capsule_finalizer
else:
    _capsule_finalizer_c = None


class CapsuleBase:

    _cdata: FFI.CData

    @property
    def typeof(self) -> int:
        return _TYPEOF(self._cdata)

    @property
    def rid(self) -> int:
        return get_rid(self._cdata)


class UninitializedRCapsule(CapsuleBase):

    @property
    def _cdata(self):
        raise RuntimeError('The embedded R is not initialized.')

    @property
    def typeof(self) -> int:
        return self._typeof

    def __init__(self, typeof):
        self._typeof = typeof


class UnmanagedSexpCapsule(CapsuleBase):

    __slots__ = ('_cdata', )

    def __init__(self, cdata: FFI.CData):
        assert is_cdata_sexp(cdata)
        self._cdata = cdata


class SexpCapsule(CapsuleBase):

    __slots__ = ('_cdata', )

    def __init__(self, cdata: FFI.CData):
        assert is_cdata_sexp(cdata)
        _preserve(cdata)
        self._cdata = cdata

    def __del__(self):
        try:
            _release(self._cdata)
        except Exception as e:
            # _release can be None while capsules when Python is terminating
            # and R is being shutdown, resulting in a race condition when
            # freeing python objects.
            if _release is None:
                pass
            else:
                raise e


class SexpCapsuleWithPassenger(SexpCapsule):
    __slots__ = ('_cdata', '_passenger')

    def __init__(self, cdata: FFI.CData, passenger, ptr):
        assert is_cdata_sexp(cdata)
        addr = _preserve(cdata)
        _PY_PASSENGER[addr] = passenger
        self._cdata = cdata
        self._passenger = ptr

    def __del__(self):
        addr = get_rid(self._cdata)
        _release(self._cdata)
        if addr not in _PY_PASSENGER:
            del(_PY_PASSENGER[addr])


class SupportsSEXP(object, metaclass=abc.ABCMeta):

    @abc.abstractproperty
    def __sexp__(self):
        pass


def _findvar(symbol, r_environment):
    rlib = openrlib.rlib
    return rlib.Rf_findVar(symbol,
                           r_environment)


def _findfun(symbol, r_environment):
    rlib = openrlib.rlib
    return rlib.Rf_findFun(symbol,
                           r_environment)


@ffi_proxy.callback(ffi_proxy._exec_findvar_in_frame_def,
                    openrlib._rinterface_cffi)
def _exec_findvar_in_frame(cdata):
    cdata_struct = openrlib.ffi.cast('struct RPY2_sym_env_data *', cdata)
    res = openrlib.rlib.Rf_findVarInFrame(
        cdata_struct.environment,
        cdata_struct.symbol
    )
    cdata_struct.data = res
    return


def findvar_in_frame(rho, symbol):
    """Safer wrapper around Rf_findVarInFrame()

    Run the function Rf_findVarInFrame() in R's C-API through
    R_ToplevelExec().

    Note: All arguments, and the object returned, are C-level
    R objects.

    Args:
    - rho: An R environment.
    - symbol: An R symbol (as returned by Rf_install())
    Returns:
    The object found.
    """
    # One would expect this to be like
    #   res = _rinterface._findfun(symbol, self.__sexp__._cdata)
    # but R's findfun will segfault if an error occurs while
    # accessing the matching object in the environment.
    exec_data = ffi.new(
        'struct RPY2_sym_env_data *',
        [symbol, rho, openrlib.rlib.R_NilValue]
    )
    _ = openrlib.rlib.R_ToplevelExec(
        openrlib.rlib._exec_findvar_in_frame,
        exec_data
    )
    if _ != openrlib.rlib.TRUE:
        raise embedded.RRuntimeError(
            'R C-API Rf_findVarInFrame()'
        )
    return exec_data.data


if FFI_MODE is ffi_proxy.InterfaceType.ABI:
    findvar_in_frame_wrap = openrlib.rlib.Rf_findVarInFrame
elif FFI_MODE is ffi_proxy.InterfaceType.API:
    findvar_in_frame_wrap = findvar_in_frame
else:
    raise ImportError('cffi mode unknown: %s' % FFI_MODE)


def _GET_DIM(robj):
    res = openrlib.rlib.Rf_getAttrib(robj,
                                     openrlib.rlib.R_DimSymbol)
    return res


def _GET_DIMNAMES(robj) -> FFI.CData:
    res = openrlib.rlib.Rf_getAttrib(robj,
                                     openrlib.rlib.R_DimNames)
    return res


def _GET_LEVELS(robj: FFI.CData) -> FFI.CData:
    res = openrlib.rlib.Rf_getAttrib(robj,
                                     openrlib.rlib.R_LevelsSymbol)
    return res


def _GET_NAMES(robj: FFI.CData) -> FFI.CData:
    res = openrlib.rlib.Rf_getAttrib(robj,
                                     openrlib.rlib.R_NamesSymbol)
    return res


def _TYPEOF(robj: FFI.CData) -> int:
    return robj.sxpinfo.type


def _SET_TYPEOF(robj: FFI.CData, v: int):
    robj.sxpinfo.type = v


def _NAMED(robj: FFI.CData) -> int:
    return robj.sxpinfo.named


def _string_getitem(cdata: FFI.CData, i: int) -> typing.Optional[str]:
    elt = openrlib.rlib.STRING_ELT(cdata, i)
    if elt == openrlib.rlib.R_NaString:
        res = None
    else:
        res = conversion._cchar_to_str(
            openrlib.rlib.R_CHAR(elt),
            conversion._R_ENC_PY[openrlib.rlib.Rf_getCharCE(elt)]
        )
    return res


# TODO: still used ?
def _string_setitem(cdata: FFI.CData, i: int, value_b) -> None:
    rlib = openrlib.rlib
    rlib.SET_STRING_ELT(
        cdata, i, rlib.Rf_mkCharCE(value_b, conversion._CE_DEFAULT_VALUE)
    )


def _has_slot(cdata: FFI.CData, name_b) -> bool:
    res = openrlib.rlib.R_has_slot(cdata, name_b)
    return bool(res)


def build_rcall(rfunction,
                args=[],
                kwargs=[]):
    rlib = openrlib.rlib
    with memorymanagement.rmemory() as rmemory:
        rcall = rmemory.protect(
            rlib.Rf_allocList(len(args)+len(kwargs)+1)
        )
        _SET_TYPEOF(rcall, rlib.LANGSXP)
        rlib.SETCAR(rcall, rfunction)
        item = rlib.CDR(rcall)
        for val in args:
            cdata = rmemory.protect(conversion._get_cdata(val))
            rlib.SETCAR(item, cdata)
            item = rlib.CDR(item)
        for key, val in kwargs:
            if key is not None:
                _assert_valid_slotname(key)
                rlib.SET_TAG(
                    item,
                    rlib.Rf_install(conversion._str_to_cchar(key)))
            cdata = rmemory.protect(conversion._get_cdata(val))
            rlib.SETCAR(item, cdata)
            item = rlib.CDR(item)
    return rcall


def _evaluated_promise(function):
    def _(*args, **kwargs):
        robj = function(*args, **kwargs)
        if _TYPEOF(robj) == openrlib.rlib.PROMSXP:
            robj = openrlib.rlib.Rf_eval(
                robj,
                openrlib.rlib.R_GlobalEnv)
        return robj
    return _


def _python_index_to_c(cdata: FFI.CData, i: int) -> int:
    """Compute the C value for a Python-style index.

    The function will translate a Python-style index that
    can be either positive or negative, if the later to
    indicate that indexing is relative to the end of the
    sequence, into a strictly positive or null index as
    use in C.

    Raises an exception IndexError if out of bounds."""

    size = openrlib.rlib.Rf_xlength(cdata)
    if i < 0:
        # x = [0,1,2,3]
        # x[-3] = x[size + (-3)] = x[4 - 3] = x[1] = 1
        # x[-2] = x[size + (-2)] = x[4 - 2] = x[2] = 2
        i = size + i
    if i >= size or i < 0:
        raise IndexError('index out of range')
    return i


# TODO: make it a general check for names or symbols ?
def _assert_valid_slotname(name: str) -> None:
    if not isinstance(name, str):
        raise ValueError('Invalid name %s' % repr(name))
    elif len(name) == 0:
        raise ValueError('The name cannot be an empty string')


def _list_attrs(cdata: FFI.CData) -> FFI.CData:
    rlib = openrlib.rlib
    attrs = rlib.ATTRIB(cdata)
    nvalues = rlib.Rf_length(attrs)
    if rlib.Rf_isList(cdata):
        namesattr = rlib.Rf_getAttrib(cdata, rlib.R_NamesSymbol)
        if namesattr != rlib.R_NilValue:
            nvalues += 1
    else:
        namesattr = rlib.R_NilValue

    with memorymanagement.rmemory() as rmemory:
        names = rmemory.protect(
            rlib.Rf_allocVector(rlib.STRSXP, nvalues))
        attr_i = 0
        if namesattr != rlib.R_NilValue:
            rlib.SET_STRING_ELT(names, attr_i,
                                rlib.PRINTNAME(rlib.R_NamesSymbol))
            attr_i += 1

        while attrs != rlib.R_NilValue:
            tag = rlib.TAG(attrs)
            if _TYPEOF(tag) == rlib.SYMSXP:
                rlib.SET_STRING_ELT(names, attr_i,
                                    rlib.PRINTNAME(tag))
            else:
                rlib.SET_STRING_ELT(
                    names,
                    attr_i,
                    rlib.R_BlankString
                )
            attrs = rlib.CDR(attrs)
            attr_i += 1
    return names


def _remove(name: FFI.CData, env: FFI.CData, inherits) -> FFI.CData:
    rlib = openrlib.rlib
    with memorymanagement.rmemory() as rmemory:
        internal = rmemory.protect(
            rlib.Rf_install(conversion._str_to_cchar('.Internal')))
        remove = rmemory.protect(
            rlib.Rf_install(conversion._str_to_cchar('remove')))
        args = rmemory.protect(rlib.Rf_lang4(remove, name,
                                             env, inherits))
        call = rmemory.protect(rlib.Rf_lang2(internal, args))
        # TODO: use tryEval instead and catch errors.
        res = rlib.Rf_eval(call, rlib.R_GlobalEnv)
    return res


def _geterrmessage() -> str:
    rlib = openrlib.rlib
    # TODO: use isString and installTrChar
    with memorymanagement.rmemory() as rmemory:
        symbol = rmemory.protect(
            rlib.Rf_install(
                conversion._str_to_cchar('geterrmessage'))
        )
        geterrmessage = _findvar(
            symbol,
            rlib.R_GlobalEnv)
        call_r = rlib.Rf_lang1(geterrmessage)
        res = rmemory.protect(
            rlib.Rf_eval(call_r, rlib.R_GlobalEnv)
        )
        res = _string_getitem(res, 0)
    return res


def serialize(cdata: FFI.CData, cdata_env: FFI.CData) -> FFI.CData:
    """Serialize an R object to an R string.

    Note that the R string returned is *not* protected from
    the R garbage collection."""

    rlib = openrlib.rlib

    with memorymanagement.rmemory() as rmemory:
        sym_serialize = rmemory.protect(
            rlib.Rf_install(conversion._str_to_cchar('serialize'))
        )
        func_serialize = rmemory.protect(
            _findfun(sym_serialize, rlib.R_BaseEnv))
        r_call = rmemory.protect(
            rlib.Rf_lang3(func_serialize, cdata, rlib.R_NilValue)
        )
        error_occured = ffi.new('int *', 0)
        res = rlib.R_tryEval(r_call,
                             cdata_env,
                             error_occured)
        if error_occured[0]:
            raise embedded.RRuntimeError(_geterrmessage())
        return res


def unserialize(cdata: FFI.CData, cdata_env: FFI.CData) -> FFI.CData:
    """Unserialize an R string to an R object.

    Note that the R object returned is *not* protected from
    the R garbage collection."""

    rlib = openrlib.rlib

    with memorymanagement.rmemory() as rmemory:
        sym_unserialize = rmemory.protect(
            rlib.Rf_install(
                conversion._str_to_cchar('unserialize'))
        )
        func_unserialize = rmemory.protect(_findfun(sym_unserialize,
                                                    rlib.R_BaseEnv))
        r_call = rmemory.protect(
            rlib.Rf_lang2(func_unserialize, cdata)
        )
        error_occured = ffi.new('int *', 0)
        res = rlib.R_tryEval(r_call,
                             cdata_env,
                             error_occured)
        if error_occured[0]:
            raise embedded.RRuntimeError(_geterrmessage())

        return res


@ffi_proxy.callback(ffi_proxy._evaluate_in_r_def,
                    openrlib._rinterface_cffi)
def _evaluate_in_r(rargs: FFI.CData) -> FFI.CData:
    # An uncaught exception in the boby of this function would
    # result in a segfault. we wrap it in a try-except an report
    # exceptions as logs.

    rlib = openrlib.rlib

    try:
        rargs = rlib.CDR(rargs)
        cdata = rlib.CAR(rargs)
        if (_TYPEOF(cdata) != rlib.EXTPTRSXP):
            # TODO: also check tag
            #    (rlib.R_ExternalPtrTag(sexp) == '.Python')
            logger.error('The fist item is not an R external pointer.')
            return rlib.R_NilValue
        handle = rlib.R_ExternalPtrAddr(cdata)
        func = ffi.from_handle(handle)

        pyargs = []
        pykwargs = {}
        rargs = rlib.CDR(rargs)
        while rargs != rlib.R_NilValue:
            cdata = rlib.CAR(rargs)
            if rlib.Rf_isNull(rlib.TAG(rargs)):
                # Unnamed argument
                pyargs.append(conversion._cdata_to_rinterface(cdata))
            else:
                # Named arguments
                rname = rlib.PRINTNAME(rlib.TAG(rargs))
                name = conversion._cchar_to_str(
                    rlib.R_CHAR(rname),
                    conversion._R_ENC_PY[openrlib.rlib.Rf_getCharCE(rname)]
                )
                pykwargs[name] = conversion._cdata_to_rinterface(cdata)
            rargs = rlib.CDR(rargs)

        res = func(*pyargs, **pykwargs)
        # The object is whatever the "rternalized" function `func`
        # is returning and we need to cast that result into a SEXP
        # that R's C API can handle. At the same time we need to ensure
        # that the R is:
        # - protected from garbage collection long enough to let the R
        #   code that called the rternalized function complete.
        # - eventually its memory is freed to prevent a leak.
        # To that end, we create a SEXP object to be returned that is
        # not managed by rpy2, leaving the object's lifespan under R's
        # sole control.
        if (
                hasattr(res, '_sexpobject') and
                isinstance(res._sexpobject, SexpCapsule)
        ):
            return res._sexpobject._cdata
        else:
            return conversion._python_to_cdata(res)
    except Exception as e:
        logger.error('%s: %s' % (type(e), e))
        return rlib.R_NilValue


def _register_external_symbols() -> None:
    python_cchar = ffi.new('char []', b'.Python')
    ffi_proxy = openrlib.ffi_proxy
    if (
            ffi_proxy.get_ffi_mode(openrlib._rinterface_cffi)
            ==
            ffi_proxy.InterfaceType.ABI
    ):
        python_callback = ffi.cast('DL_FUNC', _evaluate_in_r)
    else:
        python_callback = ffi.cast('DL_FUNC', openrlib.rlib._evaluate_in_r)
    externalmethods = ffi.new(
        'R_ExternalMethodDef[]',
        [[python_cchar,
          python_callback,
          -1],
         [ffi.NULL, ffi.NULL, 0]])
    openrlib.rlib.R_registerRoutines(
        openrlib.rlib.R_getEmbeddingDllInfo(),
        ffi.NULL,
        ffi.NULL,
        ffi.NULL,
        externalmethods
    )


class PARSING_STATUS(enum.Enum):
    PARSE_NULL = openrlib.rlib.PARSE_NULL
    PARSE_OK = openrlib.rlib.PARSE_OK
    PARSE_INCOMPLETE = openrlib.rlib.PARSE_INCOMPLETE
    PARSE_ERROR = openrlib.rlib.PARSE_ERROR
    PARSE_EOF = openrlib.rlib.PARSE_EOF


class RParsingError(Exception):

    def __init__(self, msg: str, status: PARSING_STATUS = None):
        full_msg = (
            '{msg} - {status}'
            .format(msg=msg, status=status)
        )
        super().__init__(full_msg)
        self.status = status


@ffi_proxy.callback(ffi_proxy._parsevector_wrap_def,
                    openrlib._rinterface_cffi)
def _parsevector_wrap(data: FFI.CData):
    try:
        cdata, num, status = ffi.from_handle(data)
        res = openrlib.rlib.R_ParseVector(
            cdata,  # text
            num,
            status,
            openrlib.rlib.R_NilValue)
    except Exception as e:
        res = openrlib.rlib.R_NilValue
        logger.error('_parsevector_wrap: %s', str(e))
    return res


# TODO: is this complete ?
@ffi_proxy.callback(ffi_proxy._handler_def,
                    openrlib._rinterface_cffi)
def _handler_wrap(cond, hdata):
    return openrlib.rlib.R_NilValue


if FFI_MODE is ffi_proxy.InterfaceType.ABI:
    _parsevector_wrap = _parsevector_wrap
    _handler_wrap = _handler_wrap
elif FFI_MODE is ffi_proxy.InterfaceType.API:
    _parsevector_wrap = openrlib.rlib._parsevector_wrap
    _handler_wrap = openrlib.rlib._handler_wrap
else:
    raise ImportError('cffi mode unknown: %s' % FFI_MODE)


def _parse(cdata: FFI.CData, num, rmemory) -> FFI.CData:
    status = ffi.new('ParseStatus[1]', None)
    data = ffi.new_handle((cdata, num, status))
    hdata = ffi.NULL
    res = rmemory.protect(
        openrlib.rlib.R_tryCatchError(
            _parsevector_wrap, data,
            _handler_wrap, hdata
        )
    )
    # TODO: design better handling of possible status:
    # PARSE_NULL,
    # PARSE_OK,
    # PARSE_INCOMPLETE,
    # PARSE_ERROR,
    # PARSE_EOF
    if status[0] != openrlib.rlib.PARSE_OK:
        raise RParsingError('Parsing status not OK',
                            status=PARSING_STATUS(status[0]))
    return res



import pytest
import types
import warnings
from rpy2.robjects import vectors
from rpy2.robjects.packages import importr

try:
    import IPython
except ModuleNotFoundError as no_ipython:
    warnings.warn(str(no_ipython))
    IPython = None

if IPython is None:
    html = types.SimpleNamespace(html_vector_horizontal=None,
                                 html_rlist=None,
                                 html_rdataframe=None,
                                 html_sourcecode=None,
                                 html_ridentifiedobject=None)
else:
    from rpy2.ipython import html
    
base = importr('base')

@pytest.mark.skipif(IPython is None,
                    reason='The optional package IPython cannot be imported.')
@pytest.mark.parametrize(
    'o,func',
    [(vectors.IntVector([1, 2, 3]), html.html_vector_horizontal),
     (vectors.FloatVector([1, 2, 3]), html.html_vector_horizontal),
     (vectors.StrVector(['a', 'b' 'c']), html.html_vector_horizontal),
     (vectors.FactorVector(['a', 'b' 'c']), html.html_vector_horizontal),
     (vectors.ListVector({'a': 1, 'b': 2}), html.html_rlist),
     (vectors.DataFrame({'a': 1, 'b': 'z'}), html.html_rdataframe),
     ('x <- c(1, 2, 3)', html.html_sourcecode),
     (base.c, html.html_ridentifiedobject)])
def test_html_func(o, func):
    res = func(o)
    assert isinstance(res, str)



import pytest
import contextlib
import os
import tempfile

from rpy2.robjects.lib import grid
from rpy2.robjects.vectors import FloatVector, StrVector


# TODO: is there anything to test ?
def test_unit():
    u = grid.unit(1, 'cm')
    u = grid.unit(x=1, units='cm')


# TODO: is there anything to test ?
def test_gpar():
    g = grid.gpar(col='blue')


# TODO: is there anything to test ?
def test_grob():
    g = grid.grob(name='foo')


# TODO: is there anything to test ?
def test_rect():
    r = grid.rect(x=grid.unit(0, 'cm'))


# TODO: is there anything to test ?
def test_lines():
    l = grid.lines(x=grid.unit(FloatVector([0, 1]), 'cm'))


# TODO: is there anything to test ?
def test_circle():
    c = grid.circle(x=0)


# TODO: is there anything to test ?
def test_points():
    c = grid.points()


# TODO: is there anything to test ?
def test_text():
    t = grid.text('t0')


# TODO: is there anything to test ?
def test_gtree():
    g = grid.GTree.gtree()


# TODO: is there anything to test ?
def test_grobtree():
    g = grid.GTree.grobtree()


# TODO: is there anything to test ?
@pytest.mark.parametrize('axis', (grid.XAxis.xaxis, grid.YAxis.yaxis))
def test_axis(axis):
    g = axis()


# TODO: is there anything to test ?
@pytest.mark.parametrize('axisgrob', (grid.XAxis.xaxisgrob, grid.YAxis.yaxisgrob))
def test_axisgrob(axisgrob):
    g = axisgrob()


# TODO: is there anything to test ?
def test_viewport():
    v = grid.viewport()



"""This module runs continuous updates for R, such as redrawing graphs when
the plot window is resized. Use the start() and stop() functions to turn
updates on and off.

"""
from rpy2.rinterface import process_revents
import time
import warnings
import threading


class _EventProcessorThread(threading.Thread):
    """ Call rinterface.process_revents(), pausing 
    for at least EventProcessor.interval between calls. """
    _continue = True

    def run(self):
        while self._continue:
            process_revents()
            time.sleep(EventProcessor.interval)
            
class EventProcessor(object):
    """ Processor for R events (Singleton class) """
    interval = 0.2
    daemon_thread = True
    name_thread = 'rpy2_process_revents'
    _thread = None
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = object.__new__(cls)
        return cls._instance
    
    def start(self):
        """ start the event processor """
        if (self._thread is not None) and (self._thread.is_alive()):
            raise warnings.warn("Processing of R events already started.")
        else:
            self._thread = _EventProcessorThread(name = self.name_thread)
            self._thread.setDaemon(self.daemon_thread)
            self._thread.start()

    def stop(self):
        """ stop the event processor """
        self._thread._continue = False
        self._thread.join()
    
    thread = property(lambda self: self._thread, 
                      None, None, "Thread that processes the events.")

def start():
    """ Start the threaded processing of R events. """
    EventProcessor().start()

def stop():
    """ Stop the threaded processing of R events. """
    EventProcessor().stop()




def default_key(x):
    """ Default comparison function."""
    return x


def order(seq, key=default_key, reverse=False):
    """ Return the order in which to take the items to obtained
    a sorted sequence."""
    o = list(range(len(seq)))

    def wrap_key(x):
        x = seq[x]
        return key(x)

    o.sort(key=wrap_key, reverse=reverse)

    return o



"""Callbacks available from R's C-API.

The callbacks make available in R's C-API can be specified as
Python functions, with the module providing the adapter code
that makes it possible."""

from contextlib import contextmanager
import logging
import typing
from rpy2.rinterface_lib import openrlib
from rpy2.rinterface_lib import ffi_proxy
from rpy2.rinterface_lib import conversion

logger = logging.getLogger(__name__)

_CCHAR_ENCODING = 'utf-8'


# TODO: rename to "replace_in_module"
@contextmanager
def obj_in_module(module, name: str, obj: typing.Any):
    obj_orig = getattr(module, name)
    setattr(module, name, obj)
    try:
        yield
    finally:
        setattr(module, name, obj_orig)


def consoleflush():
    pass


_FLUSHCONSOLE_EXCEPTION_LOG = 'R[flush console]: %s'


@ffi_proxy.callback(ffi_proxy._consoleflush_def,
                    openrlib._rinterface_cffi)
def _consoleflush() -> None:
    try:
        consoleflush()
    except Exception as e:
        logger.error(_FLUSHCONSOLE_EXCEPTION_LOG, str(e))


def consoleread(prompt: str) -> str:
    """Read input for the R console.

    :param prompt: The message prompted.
    :return: A string with the input returned by the user.
    """
    return input(prompt)


_READCONSOLE_EXCEPTION_LOG = 'R[read into console]: %s'
_READCONSOLE_INTERNAL_EXCEPTION_LOG = ('Internal rpy2 error with '
                                       '_consoleread callback: %s')


@ffi_proxy.callback(ffi_proxy._consoleread_def,
                    openrlib._rinterface_cffi)
def _consoleread(prompt, buf, n: int, addtohistory) -> int:
    success = None
    try:
        s = conversion._cchar_to_str(prompt, _CCHAR_ENCODING)
        reply = consoleread(s)
    except Exception as e:
        success = 0
        logger.error(_READCONSOLE_EXCEPTION_LOG, str(e))
    if success == 0:
        return success

    try:
        # TODO: Should the coding be dynamically extracted from
        # elsewhere ?
        reply_b = reply.encode('utf-8')
        reply_n = min(n, len(reply_b))
        pybuf = bytearray(n)
        pybuf[:reply_n] = reply_b[:reply_n]
        pybuf[reply_n] = ord('\n')
        pybuf[reply_n+1] = 0
        openrlib.ffi.memmove(buf,
                             pybuf,
                             n)
        if reply_n == 0:
            success = 0
        else:
            success = 1
    except Exception as e:
        success = 0
        logger.error(_READCONSOLE_INTERNAL_EXCEPTION_LOG,
                     str(e))
    return success


def consolereset() -> None:
    pass


_RESETCONSOLE_EXCEPTION_LOG = 'R[reset console]: %s'


@ffi_proxy.callback(ffi_proxy._consolereset_def,
                    openrlib._rinterface_cffi)
def _consolereset() -> None:
    try:
        consolereset()
    except Exception as e:
        logger.error(_RESETCONSOLE_EXCEPTION_LOG, str(e))


def consolewrite_print(s: str) -> None:
    """R writing to the console/terminal.

    :param s: the data to write to the console/terminal.
    """
    # TODO: is the callback for flush working with Linux ?
    print(s, end='', flush=True)


def consolewrite_warnerror(s: str) -> None:
    # TODO: use an rpy2/R-specific warning instead of UserWarning.
    logger.warning(_WRITECONSOLE_EXCEPTION_LOG, s)


_WRITECONSOLE_EXCEPTION_LOG = 'R[write to console]: %s'


@ffi_proxy.callback(ffi_proxy._consolewrite_ex_def,
                    openrlib._rinterface_cffi)
def _consolewrite_ex(buf, n: int, otype: int) -> None:
    s = conversion._cchar_to_str_with_maxlen(buf, n, _CCHAR_ENCODING)
    try:
        if otype == 0:
            consolewrite_print(s)
        else:
            consolewrite_warnerror(s)
    except Exception as e:
        logger.error(_WRITECONSOLE_EXCEPTION_LOG, str(e))


def showmessage(s: str) -> None:
    print('R wants to show a message')
    print(s)


_SHOWMESSAGE_EXCEPTION_LOG = 'R[show message]: %s'


@ffi_proxy.callback(ffi_proxy._showmessage_def,
                    openrlib._rinterface_cffi)
def _showmessage(buf):
    s = conversion._cchar_to_str(buf, _CCHAR_ENCODING)
    try:
        showmessage(s)
    except Exception as e:
        logger.error(_SHOWMESSAGE_EXCEPTION_LOG, str(e))


def choosefile(new):
    return input('Enter file name:')


_CHOOSEFILE_EXCEPTION_LOG = 'R[choose file]: %s'


@ffi_proxy.callback(ffi_proxy._choosefile_def,
                    openrlib._rinterface_cffi)
def _choosefile(new, buf, n: int) -> int:
    try:
        res = choosefile(new)
    except Exception as e:
        logger.error(_CHOOSEFILE_EXCEPTION_LOG, str(e))
        res = None

    if res is None:
        return 0

    res_cdata = conversion._str_to_cchar(res)
    openrlib.ffi.memmove(buf, res_cdata, len(res_cdata))
    return len(res_cdata)


def showfiles(filenames: typing.Tuple[str, ...],
              headers: typing.Tuple[str, ...],
              wtitle: typing.Optional[str],
              pager: typing.Optional[str]) -> None:
    """R showing files.

    :param filenames: A tuple of file names.
    :param headers: A tuple of strings (TODO: check what it is)
    :wtitle: Title of the "window" showing the files.
    :pager: Pager to use to show the list of files.
    """
    for fn in filenames:
        print('File: %s' % fn)
        with open(fn) as fh:
            for row in fh:
                print(row)
            print('---')


_SHOWFILE_EXCEPTION_LOG = 'R[show file]: %s'
_SHOWFILE_INTERNAL_EXCEPTION_LOG = ('Internal rpy2 error while '
                                    'showing files for R: %s')


@ffi_proxy.callback(ffi_proxy._showfiles_def,
                    openrlib._rinterface_cffi)
def _showfiles(nfiles: int, files, headers, wtitle, delete, pager) -> int:
    filenames = []
    headers_str = []
    wtitle_str = None
    pager_str = None
    try:
        wtitle_str = conversion._cchar_to_str(wtitle, _CCHAR_ENCODING)
        pager_str = conversion._cchar_to_str(pager, _CCHAR_ENCODING)
        for i in range(nfiles):
            filenames.append(
                conversion._cchar_to_str(files[i],
                                         _CCHAR_ENCODING)
            )
            headers_str.append(
                conversion._cchar_to_str(headers[i],
                                         _CCHAR_ENCODING)
            )
    except Exception as e:
        logger.error(_SHOWFILE_INTERNAL_EXCEPTION_LOG, str(e))

    if len(filenames):
        res = 0
    else:
        res = 1
    try:
        showfiles(tuple(filenames),
                  tuple(headers_str),
                  wtitle_str,
                  pager_str)
    except Exception as e:
        res = 1
        logger.error(_SHOWFILE_EXCEPTION_LOG, str(e))

    return res


def cleanup(saveact, status, runlast):
    pass


_CLEANUP_EXCEPTION_LOG = 'R[cleanup]: %s'


@ffi_proxy.callback(ffi_proxy._cleanup_def,
                    openrlib._rinterface_cffi)
def _cleanup(saveact, status, runlast):
    try:
        cleanup(saveact, status, runlast)
    except Exception as e:
        logger.error(_CLEANUP_EXCEPTION_LOG, str(e))


def processevents() -> None:
    """Process R events.

    This function can be periodically called by R to handle
    events such as window resizing in an interactive graphical
    device."""
    pass


_PROCESSEVENTS_EXCEPTION_LOG = 'R[processevents]: %s'


@ffi_proxy.callback(ffi_proxy._processevents_def,
                    openrlib._rinterface_cffi)
def _processevents() -> None:
    try:
        processevents()
    except Exception as e:
        logger.error(_PROCESSEVENTS_EXCEPTION_LOG, str(e))


def busy(x: int) -> None:
    """R is busy.

    :param x: TODO this is an integer but I do not know what it does.
    """
    pass


_BUSY_EXCEPTION_LOG = 'R[busy]: %s'


@ffi_proxy.callback(ffi_proxy._busy_def,
                    openrlib._rinterface_cffi)
def _busy(which: int) -> None:
    try:
        busy(which)
    except Exception as e:
        logger.error(_BUSY_EXCEPTION_LOG, str(e))


def callback() -> None:
    pass


_CALLBACK_EXCEPTION_LOG = 'R[callback]: %s'


@ffi_proxy.callback(ffi_proxy._callback_def,
                    openrlib._rinterface_cffi)
def _callback() -> None:
    try:
        callback()
    except Exception as e:
        logger.error(_CALLBACK_EXCEPTION_LOG, str(e))


def yesnocancel(question: str) -> int:
    """Asking a user to answer yes, no, or cancel.

    :param question: The question asked to the user
    :return: An integer with the answer.
    """
    return int(input(question))


_YESNOCANCEL_EXCEPTION_LOG = 'R[yesnocancel]: %s'


@ffi_proxy.callback(ffi_proxy._yesnocancel_def,
                    openrlib._rinterface_cffi)
def _yesnocancel(question):
    try:
        q = conversion._cchar_to_str(question, _CCHAR_ENCODING)
        res = yesnocancel(q)
    except Exception as e:
        logger.error(_YESNOCANCEL_EXCEPTION_LOG, str(e))
    return res



from contextlib import contextmanager


@contextmanager
def obj_in_module(module, name, obj):
    backup_obj = getattr(module, name, None)
    setattr(module, name, obj)
    try:
        yield
    finally:
        if backup_obj:
            setattr(module, name, backup_obj)

def assert_equal_sequence(x, y):
    assert type(x) is type(y)
    assert len(x) == len(y)
    assert all(x_e == y_e for x_e, y_e in zip(x, y))


def noconsole(x):
    pass



# coding: utf-8
import pytest
import rpy2.rinterface as rinterface
import rpy2.rlike.container as rlc

rinterface.initr()


def _noconsole(x):
    pass


@pytest.fixture(scope='module')
def silent_consolewrite():
    module = rinterface.callbacks
    name = 'consolewrite_print'
    backup_func = getattr(module, name)
    setattr(module, name, _noconsole)
    try:
        yield
    finally:
        setattr(module, name, backup_func)


def test_new():
    x = 'a'
    with pytest.raises(ValueError):
        rinterface.SexpClosure(x)


def test_typeof():
    sexp = rinterface.globalenv.find('plot')
    assert sexp.typeof == rinterface.RTYPES.CLOSXP


def test_r_error():
    r_sum = rinterface.baseenv['sum']
    letters = rinterface.baseenv['letters']
    with pytest.raises(rinterface.embedded.RRuntimeError),\
            pytest.warns(rinterface.RRuntimeWarning):
        r_sum(letters)


def test_string_argument():
    r_nchar = rinterface.baseenv['::']('base', 'nchar')
    res = r_nchar('foo')
    assert res[0] == 3


def test_utf8_argument_name():
    c = rinterface.globalenv.find('c')
    d = dict([(u'哈哈', 1)])
    res = c(**d)
    assert u'哈哈' == res.do_slot('names')[0]


def test_emptystringparams():
    d = dict([('', 1)])
    with pytest.raises(ValueError):
        rinterface.baseenv['list'](**d)


def test_closureenv_isenv():
    exp = rinterface.parse('function() { }')
    fun = rinterface.baseenv['eval'](exp)
    assert isinstance(fun.closureenv, rinterface.SexpEnvironment)


def test_closureenv():

    assert 'y' not in rinterface.globalenv

    exp = rinterface.parse('function(x) { x[y] }')
    fun = rinterface.baseenv['eval'](exp)
    vec = rinterface.baseenv['letters']

    assert isinstance(fun.closureenv, rinterface.SexpEnvironment)

    with pytest.raises(rinterface.embedded.RRuntimeError):
        with pytest.warns(rinterface.RRuntimeWarning):
            fun(vec)

    fun.closureenv['y'] = (rinterface
                           .IntSexpVector([1, ]))
    assert 'a' == fun(vec)[0]

    fun.closureenv['y'] = (rinterface
                           .IntSexpVector([2, ]))
    assert 'b' == fun(vec)[0]


def test_call_s4_setClass():
    # R's package "methods" can perform uncommon operations
    r_setClass = rinterface.globalenv.find('setClass')
    r_representation = rinterface.globalenv.find('representation')
    attrnumeric = rinterface.StrSexpVector(['numeric', ])
    classname = rinterface.StrSexpVector(['Track', ])
    classrepr = r_representation(x=attrnumeric,
                                 y=attrnumeric)
    r_setClass(classname,
               classrepr)
    # TODO: where is the test ?


def test_call_OrdDict():
    ad = rlc.OrdDict((('a', rinterface.IntSexpVector([2, ])),
                      ('b', rinterface.IntSexpVector([1, ])),
                      (None, rinterface.IntSexpVector([5, ])),
                      ('c', rinterface.IntSexpVector([0, ]))))

    mylist = rinterface.baseenv['list'].rcall(tuple(ad.items()),
                                              rinterface.globalenv)

    names = [x for x in mylist.do_slot('names')]

    for i in range(4):
        assert ('a', 'b', '', 'c')[i] == names[i]


def test_call_OrdDictEnv():
    ad = rlc.OrdDict(((None, rinterface.parse('sum(x)')), ))
    env_a = rinterface.baseenv['new.env']()
    env_a['x'] = rinterface.IntSexpVector([1, 2, 3])
    sum_a = rinterface.baseenv['eval'].rcall(tuple(ad.items()), env_a)
    assert 6 == sum_a[0]
    env_b = rinterface.baseenv['new.env']()
    env_b['x'] = rinterface.IntSexpVector([4, 5, 6])
    sum_b = rinterface.baseenv['eval'].rcall(tuple(ad.items()), env_b)
    assert 15 == sum_b[0]


def test_error_in_call():
    r_sum = rinterface.baseenv['sum']

    with pytest.raises(rinterface.embedded.RRuntimeError),\
            pytest.warns(rinterface.RRuntimeWarning):
        r_sum(2, 'a')


def test_missing_arg():
    exp = rinterface.parse('function(x) { missing(x) }')
    fun = rinterface.baseenv['eval'](exp)
    nonmissing = rinterface.IntSexpVector([0, ])
    missing = rinterface.MissingArg
    assert not fun(nonmissing)[0]
    assert fun(missing)[0]


def test_scalar_convert_integer():
    assert 'integer' == rinterface.baseenv['typeof'](int(1))[0]


def test_scalar_convert_double():
    assert 'double' == rinterface.baseenv['typeof'](1.0)[0]


def test_scalar_convert_boolean():
    assert 'logical' == rinterface.baseenv['typeof'](True)[0]


@pytest.mark.parametrize('give_env', (True, False))
@pytest.mark.parametrize('use_rlock', (True, False))
def test_call_in_context(give_env, use_rlock):
    ls = rinterface.baseenv['ls']
    get = rinterface.baseenv['get']
    if give_env:
        env = rinterface.baseenv['new.env']()
    else:
        env = None
    assert 'foo' not in ls()
    with rinterface.local_context(env=env, use_rlock=use_rlock) as lc:
        lc['foo'] = 123
        assert tuple(get('foo')) == (123, )
    assert 'foo' not in ls()


@pytest.mark.parametrize('use_rlock', (True, False))
def test_call_in_context_nested(use_rlock):
    ls = rinterface.baseenv['ls']
    get = rinterface.baseenv['get']
    assert 'foo' not in ls()
    with rinterface.local_context() as lc_a:
        lc_a['foo'] = 123
        assert tuple(get('foo')) == (123, )
        with rinterface.local_context(use_rlock=use_rlock) as lc_b:
            lc_b['foo'] = 456
            assert tuple(get('foo')) == (456, )
        assert tuple(get('foo')) == (123, )
    assert 'foo' not in ls()



import io
import os
import pytest
import sys
import rpy2.robjects as robjects
import rpy2.robjects.help
import rpy2.robjects.packages as packages
import rpy2.robjects.packages_utils
from rpy2.rinterface_lib.embedded import RRuntimeError

rinterface = robjects.rinterface


class TestPackage(object):
    
    def tests_package_from_env(self):
        env = robjects.Environment()
        env['a'] = robjects.StrVector('abcd')
        env['b'] = robjects.IntVector((1,2,3))
        env['c'] = robjects.r(''' function(x) x^2''')
        pck = robjects.packages.Package(env, "dummy_package")
        assert isinstance(pck.a, robjects.Vector)
        assert isinstance(pck.b, robjects.Vector)
        assert isinstance(pck.c, robjects.Function)


    def test_new_with_dot(self):
        env = robjects.Environment()
        env['a.a'] = robjects.StrVector('abcd')
        env['b'] = robjects.IntVector((1,2,3))
        env['c'] = robjects.r(''' function(x) x^2''')
        pck = robjects.packages.Package(env, "dummy_package")
        assert isinstance(pck.a_a, robjects.Vector)
        assert isinstance(pck.b, robjects.Vector)
        assert isinstance(pck.c, robjects.Function)


    def test_new_with_dot_conflict(self):
        env = robjects.Environment()
        env['a.a_a'] = robjects.StrVector('abcd')
        env['a_a.a'] = robjects.IntVector((1,2,3))
        env['c'] = robjects.r(''' function(x) x^2''')
        with pytest.raises(packages.LibraryError):
            robjects.packages.Package(env, "dummy_package")


    def test_new_with_dot_conflict2(self):
        env = robjects.Environment()
        name_in_use = dir(packages.Package(env, "foo"))[0]
        env[name_in_use] = robjects.StrVector('abcd')
        with pytest.raises(packages.LibraryError):
            robjects.packages.Package(env, "dummy_package")

    def tests_package_repr(self):
        env = robjects.Environment()
        pck = robjects.packages.Package(env, "dummy_package")
        assert isinstance(repr(pck), str)


def test_signaturetranslatedanonymouspackage():
    rcode = """
    square <- function(x) {
    return(x^2)
    }
    
    cube <- function(x) {
    return(x^3)
    }
    """
    
    powerpack = packages.STAP(rcode, "powerpack")
    assert hasattr(powerpack, 'square')
    assert hasattr(powerpack, 'cube')


def test_installedstpackage_docstring():
    stats = robjects.packages.importr('stats',
                                      on_conflict='warn')
    assert stats.__doc__.startswith('Python representation of an R package.')


def test_installedstpackage_docstring_no_rname():
    stats = robjects.packages.importr('stats',
                                      on_conflict='warn')
    stats.__rname__ = None
    assert stats.__doc__.startswith(
        'Python representation of an R package.%s'
        '<No information available>' % os.linesep)


class TestImportr(object):

    def test_importr_stats(self):
        stats = robjects.packages.importr('stats',
                                          on_conflict='warn')
        assert isinstance(stats, robjects.packages.Package)

    def test_import_stats_with_libloc(self):
        path = os.path.dirname(
            robjects.packages_utils.get_packagepath('stats')
        )
        stats = robjects.packages.importr('stats', 
                                          on_conflict='warn',
                                          lib_loc = path)
        assert isinstance(stats, robjects.packages.Package)

    def test_import_stats_with_libloc_and_suppressmessages(self):
        path = os.path.dirname(
            robjects.packages_utils.get_packagepath('stats')
        )
        stats = robjects.packages.importr('stats',
                                          lib_loc=path,
                                          on_conflict='warn',
                                          suppress_messages=False)
        assert isinstance(stats, robjects.packages.Package)

    def test_import_stats_with_libloc_with_quote(self):
        path = 'coin"coin'

        with pytest.raises(robjects.packages.PackageNotInstalledError), \
             pytest.warns(UserWarning):
            Tmp_File = io.StringIO
            tmp_file = Tmp_File()
            try:
                stdout = sys.stdout
                sys.stdout = tmp_file
                robjects.packages.importr('dummy_inexistant', lib_loc=path)
            finally:
                sys.stdout = stdout
                tmp_file.close()

    def test_import_datasets(self):
        datasets = robjects.packages.importr('datasets')
        assert isinstance(datasets, robjects.packages.Package)
        assert isinstance(datasets.__rdata__, 
                          robjects.packages.PackageData)
        assert isinstance(robjects.packages.data(datasets), 
                          robjects.packages.PackageData)

    def test_datatsets_names(self):
        datasets = robjects.packages.importr('datasets')
        datasets_data = robjects.packages.data(datasets)
        datasets_names = tuple(datasets_data.names())
        assert len(datasets_names) > 0
        assert all(isinstance(x, str) for x in datasets_names)

    def test_datatsets_fetch(self):
        datasets = robjects.packages.importr('datasets')
        datasets_data = robjects.packages.data(datasets)
        datasets_names = tuple(datasets_data.names())
        assert isinstance(datasets_data.fetch(datasets_names[0]),
                          robjects.Environment)
        with pytest.raises(KeyError):
            datasets_data.fetch('foo_%s' % datasets_names[0])


def test_wherefrom():
    stats = robjects.packages.importr('stats', on_conflict='warn')
    rnorm_pack = robjects.packages.wherefrom('rnorm')
    assert rnorm_pack.do_slot('name')[0] == 'package:stats'


def test_installedpackages():
    instapacks = robjects.packages.InstalledPackages()
    res = instapacks.isinstalled('foo')
    assert res is False
    ncols = len(instapacks.colnames)
    for row in instapacks:
        assert ncols == len(row)
        

# TODO: Not really a unit test. The design of the code being
# tested (or not tested) probably need to be re-thought.
def test_parsedcode():
    rcode = '1+2'
    expression = rinterface.parse(rcode)
    pc = robjects.packages.ParsedCode(expression)
    assert isinstance(pc, type(expression))


def test_sourcecode():
    rcode = '1+2'
    expression = rinterface.parse(rcode)
    sc = robjects.packages.SourceCode(rcode)
    assert isinstance(sc.parse(), type(expression))


def test_sourcecode_as_namespace():
    rcode = '\n'.join(
        ('x <- 1+2',
         'f <- function(x) x+1')
    )
    sc = robjects.packages.SourceCode(rcode)
    foo_ns = sc.as_namespace('foo_ns')
    assert hasattr(foo_ns, 'x')
    assert hasattr(foo_ns, 'f')
    assert foo_ns.x[0] == 3
    assert isinstance(foo_ns.f, rinterface.SexpClosure)



import os
import pytest
import rpy2.robjects as robjects
import rpy2.robjects.help as rh
rinterface = robjects.rinterface

class TestPackage(object):

    def test_init(self):
        base_help = rh.Package('base')
        assert base_help.name == 'base'

    def test_repr(self):
        base_help = rh.Package('base')
        assert isinstance(repr(base_help), str)


class TestPage(object):
    
    def test_init(self):
        base_help = rh.Package('base')
        p = base_help.fetch('print')
        assert tuple(p.sections.keys())[0] == '\\title'

    def test_fetch(self):
        base_help = rh.Package('base')
        f = base_help.fetch('print')
        assert '\\title' in f.sections.keys()

    def test_to_docstring(self):
        base_help = rh.Package('base')
        p = base_help.fetch('print')
        ds = p.to_docstring()
        assert ds[:5] == 'title'

    def test_title(self):
        base_help = rh.Package('base')
        p = base_help.fetch('print')
        d = p.title()
        assert all(isinstance(x, str) for x in d)
        assert len(d) > 0

    def test_description(self):
        base_help = rh.Package('base')
        p = base_help.fetch('print')
        d = p.description()
        assert all(isinstance(x, str) for x in d)
        assert len(d) > 0

    def test_seealso(self):
        base_help = rh.Package('base')
        p = base_help.fetch('print')
        d = p.seealso()
        assert all(isinstance(x, str) for x in d)
        assert len(d) > 0

    def test_usage(self):
        base_help = rh.Package('base')
        p = base_help.fetch('print')
        d = p.usage()
        assert all(isinstance(x, str) for x in d)
        assert len(d) > 0


@pytest.mark.xfail(
    os.name == 'nt',
    reason='Windows is missing library/translations/Meta/Rd.rds file')
def test_pages():
    pages = rh.pages('plot')
    assert isinstance(pages, tuple)
    assert all(isinstance(elt, rh.Page) for elt in pages)



"""
Package to interface with R, with a focus on interactive usage
(REPL approach to code-writing), although the package will work
in non-interactive settings.

The package aims at being simple rather than exhaustively complete
(rpy2.robjects, or rpy2.rinterface, can be used if needed), providing
a comfortable experience with autocompletion-capable python consoles.
"""

from collections import OrderedDict

from rpy2.robjects.packages import _loaded_namespaces
from rpy2.robjects.vectors import IntVector, FloatVector, ComplexVector
from rpy2.robjects.vectors import Array, Matrix
from rpy2.robjects.vectors import StrVector
from rpy2.robjects.vectors import ListVector, DataFrame
from rpy2.robjects.environments import Environment
from rpy2.rinterface import NULL
from rpy2.robjects import Formula, RS4
from rpy2.robjects import methods
from rpy2.robjects import conversion
from rpy2.robjects import help as rhelp
from rpy2.robjects.language import eval

from . import process_revents as revents

from os import linesep
import re

class S4Classes(object):
    """ *Very* experimental attempt at getting the S4 classes dynamically
    mirrored. """
    __instance = None
    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = object.__new__(cls)
        return cls.__instance

    def __setattr__(self, name, value):
        raise AttributeError("Attributes cannot be set. Use 'importr'")



#classes = S4Classes()

#revents.start()

